import asyncio
import hashlib
import shutil
import sys
import tempfile
import uuid
from collections.abc import Callable, Coroutine
from datetime import UTC, datetime
from io import BytesIO
from pathlib import Path
from shutil import which
from types import NoneType
from typing import Any, TypeVar, overload

import httpx
from adbutils import AdbClient
from dotenv import load_dotenv
from langchain_core.messages import AIMessage
from limrun_api import AsyncLimrun
from PIL import Image
from pydantic import BaseModel

from minitap.mobile_use.agents.outputter.outputter import outputter
from minitap.mobile_use.agents.planner.types import Subgoal
from minitap.mobile_use.clients.browserstack_client import BrowserStackClientWrapper
from minitap.mobile_use.clients.idb_client import IdbClientWrapper
from minitap.mobile_use.clients.ios_client import DeviceType, IosClientWrapper, get_ios_client
from minitap.mobile_use.clients.ui_automator_client import UIAutomatorClient
from minitap.mobile_use.clients.wda_client import WdaClientWrapper
from minitap.mobile_use.config import AgentNode, OutputConfig, record_events, settings
from minitap.mobile_use.context import (
    DeviceContext,
    DevicePlatform,
    ExecutionSetup,
    IsReplan,
    MobileUseContext,
)
from minitap.mobile_use.controllers.limrun_controller import (
    LimrunAndroidController,
    LimrunIosController,
)
from minitap.mobile_use.controllers.platform_specific_commands_controller import get_first_device
from minitap.mobile_use.graph.graph import get_graph
from minitap.mobile_use.graph.state import State
from minitap.mobile_use.sdk.builders.agent_config_builder import get_default_agent_config
from minitap.mobile_use.sdk.builders.task_request_builder import TaskRequestBuilder
from minitap.mobile_use.sdk.constants import DEFAULT_PROFILE_NAME
from minitap.mobile_use.sdk.services.cloud_mobile import CloudMobileService
from minitap.mobile_use.sdk.services.platform import PlatformService
from minitap.mobile_use.sdk.types.agent import AgentConfig, LimrunPlatform
from minitap.mobile_use.sdk.types.exceptions import (
    AgentError,
    AgentNotInitializedError,
    AgentProfileNotFoundError,
    AgentTaskRequestError,
    CloudMobileServiceUninitializedError,
    DeviceNotFoundError,
    ExecutableNotFoundError,
    PlatformServiceUninitializedError,
    ServerStartupError,
)
from minitap.mobile_use.sdk.types.platform import TaskRunPlanResponse, TaskRunStatus
from minitap.mobile_use.sdk.types.task import (
    AgentProfile,
    PlatformTaskInfo,
    PlatformTaskRequest,
    Task,
    TaskRequest,
)
from minitap.mobile_use.services.telemetry import telemetry
from minitap.mobile_use.utils.app_launch_utils import _handle_initial_app_launch
from minitap.mobile_use.utils.logger import get_logger
from minitap.mobile_use.utils.media import (
    create_gif_from_trace_folder,
    create_steps_json_from_trace_folder,
    remove_images_from_trace_folder,
    remove_steps_json_from_trace_folder,
)
from minitap.mobile_use.utils.recorder import log_agent_thought

logger = get_logger(__name__)

TOutput = TypeVar("TOutput", bound=BaseModel | None)

load_dotenv()


class Agent:
    _config: AgentConfig
    _tasks: list[Task] = []
    _tmp_traces_dir: Path
    _initialized: bool = False
    _device_context: DeviceContext
    _adb_client: AdbClient | None
    _ui_adb_client: UIAutomatorClient | None
    _ios_client: IosClientWrapper | None
    _ios_device_type: DeviceType | None
    _current_task: asyncio.Task | None = None
    _task_lock: asyncio.Lock
    _cloud_mobile_id: str | None = None
    _limrun_instance_id: str | None = None
    _limrun_controller: Any = None

    def __init__(self, *, config: AgentConfig | None = None):
        self._config = config or get_default_agent_config()
        self._tasks = []
        self._tmp_traces_dir = Path(tempfile.gettempdir()) / "mobile-use-traces"
        self._initialized = False
        self._task_lock = asyncio.Lock()

        # Initialize platform service if API key is available in environment
        # Note: Can also be initialized later with API key at agent .init()
        if settings.MINITAP_API_KEY:
            self._platform_service = PlatformService()
            self._cloud_mobile_service = CloudMobileService()
        else:
            self._platform_service = None
            self._cloud_mobile_service = None

    async def init(
        self,
        api_key: str | None = None,
        server_restart_attempts: int = 3,
        retry_count: int = 5,
        retry_wait_seconds: int = 5,
    ):
        # Start telemetry session for SDK usage (if not already started by CLI)
        if not telemetry._session_id:
            telemetry.start_session({"source": "sdk"})

        try:
            return await self._init_internal(
                api_key=api_key,
                server_restart_attempts=server_restart_attempts,
                retry_count=retry_count,
                retry_wait_seconds=retry_wait_seconds,
            )
        except Exception as e:
            session_id = telemetry._session_id
            telemetry.capture_exception(e, {"phase": "agent_init"})
            telemetry.end_session(success=False, error=str(e))
            if session_id:
                logger.info(f"If you need support, please include this session ID: {session_id}")
            raise

    async def _init_internal(
        self,
        api_key: str | None = None,
        server_restart_attempts: int = 3,
        retry_count: int = 5,
        retry_wait_seconds: int = 5,
    ):
        if api_key:
            self._platform_service = PlatformService(api_key=api_key)
            self._cloud_mobile_service = CloudMobileService(api_key=api_key)

        # Skip initialization for cloud devices - no local setup required
        if self._config.cloud_mobile_id_or_ref:
            if not self._cloud_mobile_service:
                raise CloudMobileServiceUninitializedError()
            self._cloud_mobile_id = await self._cloud_mobile_service.resolve_cloud_mobile_id(
                cloud_mobile_id_or_ref=self._config.cloud_mobile_id_or_ref,
            )
            logger.info("Cloud device configured - skipping local initialization")
            self._initialized = True
            return True

        # Handle BrowserStack initialization
        if self._config.browserstack_config:
            logger.info("Initializing BrowserStack session...")
            self._ios_client = BrowserStackClientWrapper(config=self._config.browserstack_config)
            session_started = await self._ios_client.init_client()
            if not session_started:
                raise ServerStartupError(
                    message="Failed to create BrowserStack session. "
                    "Please check your credentials and device configuration."
                )
            self._ios_device_type = DeviceType.BROWSERSTACK
            self._adb_client = None
            self._ui_adb_client = None
            logger.success("BrowserStack session created successfully")

            self._device_context = await self._get_device_context(
                device_id="browserstack", platform=DevicePlatform.IOS
            )
            logger.info(self._device_context.to_str())
            logger.info("✅ Mobile-use agent initialized with BrowserStack.")
            self._initialized = True
            telemetry.capture_agent_initialized(
                platform=DevicePlatform.IOS.value,
                device_id="browserstack",
            )
            return True

        # Handle Limrun cloud device initialization
        # Check for pre-configured controllers first, then fall back to limrun_config
        if self._config.limrun_android_controller or self._config.limrun_ios_controller:
            return await self._init_limrun_device(
                android_controller=self._config.limrun_android_controller,
                ios_controller=self._config.limrun_ios_controller,
            )
        if self._config.limrun_config:
            return await self._init_limrun_device()

        if not which("adb") and not which("xcrun"):
            raise ExecutableNotFoundError("cli_tools")

        if self._initialized:
            logger.warning("Agent is already initialized. Skipping...")
            return True

        # Get first available device ID
        if not self._config.device_id or not self._config.device_platform:
            device_id, platform, ios_device_type = get_first_device(logger=logger)
        else:
            device_id, platform = self._config.device_id, self._config.device_platform
            ios_device_type = None  # Will be auto-detected in _init_clients

        if not device_id or not platform:
            error_msg = "No device found. Exiting."
            logger.error(error_msg)
            raise DeviceNotFoundError(error_msg)

        # Initialize clients
        self._init_clients(
            device_id=device_id,
            platform=platform,
            ios_device_type=ios_device_type,
            retry_count=retry_count,
            retry_wait_seconds=retry_wait_seconds,
        )

        # Initialize iOS client (IDB companion for simulators, WDA already running for physical)
        if self._ios_client:
            if isinstance(self._ios_client, IdbClientWrapper):
                logger.info("Starting IDB companion for iOS simulator...")
                companion_started = await self._ios_client.init_companion()
                if not companion_started:
                    raise ServerStartupError(
                        message="Failed to start IDB companion for iOS simulator. "
                        "Please ensure fb-idb is installed: https://fbidb.io/docs/installation/"
                    )
                logger.success("IDB companion started successfully")
            elif isinstance(self._ios_client, WdaClientWrapper):
                logger.info("Connecting to WebDriverAgent for physical iOS device...")
                wda_connected = await self._ios_client.init_client()
                if not wda_connected:
                    raise ServerStartupError(
                        message="Failed to connect to WebDriverAgent. "
                        "Please ensure WDA is running on your device. "
                        "See the setup instructions above."
                    )
                logger.success("WDA client connected for physical device")

        # Start necessary servers
        restart_attempt = 0
        while restart_attempt < server_restart_attempts:
            success = self._run_servers(
                device_id=device_id,
                platform=platform,
            )
            if success:
                break

            restart_attempt += 1
            if restart_attempt < server_restart_attempts:
                logger.warning(
                    f"Server start failed, attempting restart "
                    f"{restart_attempt}/{server_restart_attempts}"
                )
            else:
                error_msg = "Mobile-use servers failed to start after all restart attempts."
                logger.error(error_msg)
                raise ServerStartupError(message=error_msg)

        self._device_context = await self._get_device_context(
            device_id=device_id, platform=platform
        )
        logger.info(self._device_context.to_str())
        logger.info("✅ Mobile-use agent initialized.")
        self._initialized = True
        telemetry.capture_agent_initialized(
            platform=platform.value,
            device_id=device_id,
        )
        return True

    async def install_apk(self, apk_path: str | Path) -> None:
        """
        Install an APK on the connected device.
        For cloud mobiles, the APK must be x86_64 compatible.

        Args:
            apk_path: Path to the local APK file to install

        Raises:
            AgentNotInitializedError: If the agent is not initialized
            AgentError: If attempting to install on non-Android device or ADB operations fail
            FileNotFoundError: If the APK file doesn't exist
            CloudMobileServiceUninitializedError: If cloud service is unavailable
        """
        try:
            await self._install_apk_internal(apk_path)
        except Exception as e:
            telemetry.capture_exception(e, {"phase": "install_apk"})
            raise

    async def _install_apk_internal(self, apk_path: str | Path) -> None:
        if isinstance(apk_path, str):
            apk_path = Path(apk_path)

        if not apk_path.exists():
            raise FileNotFoundError(f"APK file not found: {apk_path}")

        if self._config.cloud_mobile_id_or_ref:
            await self._install_apk_on_cloud_mobile(apk_path)
        else:
            if not self._initialized:
                raise AgentNotInitializedError()

            if self._device_context.mobile_platform != DevicePlatform.ANDROID:
                raise AgentError(
                    "APK can only be installed on Android devices but got "
                    f"'{self._device_context.mobile_platform.value}'"
                )

            device_id = self._device_context.device_id
            logger.info(f"Installing APK on Android device '{device_id}'")
            if not self._adb_client:
                raise AgentError("ADB client not initialized")

            device = self._adb_client.device(serial=device_id)
            await asyncio.to_thread(device.install, apk_path)
            logger.info(f"APK installed successfully on Android device '{device_id}'")

    async def _install_apk_on_cloud_mobile(self, apk_path: Path) -> None:
        """
        Install an APK on a cloud mobile device.

        This method starts the cloud mobile if needed, then uploads and installs the APK.
        """
        if not self._cloud_mobile_id:
            raise AgentTaskRequestError("Cloud mobile ID is not configured")

        if not self._cloud_mobile_service:
            raise CloudMobileServiceUninitializedError()

        # Check platform before starting - fail early if not Android
        vm_info = await self._cloud_mobile_service._get_virtual_mobile_status(self._cloud_mobile_id)
        if vm_info.platform and vm_info.platform != "android":
            raise AgentError(
                f"APK can only be installed on Android cloud mobiles but got '{vm_info.platform}'"
            )

        # Start cloud mobile if not already started
        logger.info(f"Starting cloud mobile '{self._cloud_mobile_id}' for APK installation...")
        await self._cloud_mobile_service.start_and_wait_for_ready(
            cloud_mobile_id=self._cloud_mobile_id,
        )

        # Install APK
        logger.info(f"Installing APK '{apk_path.name}' on cloud mobile '{self._cloud_mobile_id}'")
        await self._cloud_mobile_service.install_apk(
            cloud_mobile_id=self._cloud_mobile_id,
            apk_path=apk_path,
        )
        logger.success(f"APK '{apk_path.name}' installed successfully")

    async def install_app(self, app_path: str | Path) -> str | None:
        """
        Install an app on the connected device.

        For Android: Installs an APK file using ADB.
        For iOS (Limrun): Uploads and installs a .app folder using diff-based
                         patch syncing for fast updates.

        Args:
            app_path: Path to the app to install:
                      - Android: Path to an APK file
                      - iOS: Path to a .app folder (simulator build)

        Returns:
            The bundle ID of the installed app (iOS only), or None for Android.

        Raises:
            AgentNotInitializedError: If the agent is not initialized
            AgentError: If installation fails or platform is unsupported
            FileNotFoundError: If the app file/folder doesn't exist
        """
        try:
            return await self._install_app_internal(app_path)
        except Exception as e:
            telemetry.capture_exception(e, {"phase": "install_app"})
            raise

    async def _install_app_internal(self, app_path: str | Path) -> str | None:
        if isinstance(app_path, str):
            app_path = Path(app_path)

        if not app_path.exists():
            raise FileNotFoundError(f"App not found: {app_path}")

        if not self._initialized:
            raise AgentNotInitializedError()

        platform = self._device_context.mobile_platform

        if platform == DevicePlatform.ANDROID:
            await self._install_apk_internal(app_path)
            return None

        elif platform == DevicePlatform.IOS:
            return await self._install_ios_app(app_path)

        else:
            raise AgentError(f"App installation not supported for platform: {platform}")

    async def _install_ios_app(self, app_path: Path) -> str:
        """
        Install an iOS .app bundle on a Limrun iOS device.

        Uses diff-based patch syncing for fast updates - only changed parts
        of the app are uploaded.

        Args:
            app_path: Path to the .app folder (simulator build)

        Returns:
            The bundle ID of the installed app

        Raises:
            AgentError: If not connected to a Limrun iOS device
            FileNotFoundError: If the .app folder doesn't exist
        """

        if not app_path.is_dir() or not app_path.suffix == ".app":
            raise AgentError(
                f"Expected a .app folder for iOS, got: {app_path}. "
                "Please provide the path to your built .app folder "
                "(e.g., build/Debug-iphonesimulator/MyApp.app)"
            )

        # Check if we have a Limrun iOS controller
        if not isinstance(self._ios_client, LimrunIosController):
            raise AgentError(
                "iOS app installation is only supported for Limrun iOS devices. "
                "For local simulators, use 'xcrun simctl install' directly."
            )

        limrun_controller: LimrunIosController = self._ios_client

        # Get API key from settings or platform service
        api_key_value: str | None = None
        if settings.MINITAP_API_KEY:
            api_key_raw = settings.MINITAP_API_KEY
            api_key_value = (
                api_key_raw.get_secret_value()
                if hasattr(api_key_raw, "get_secret_value")
                else str(api_key_raw)
            )
        if not api_key_value and self._platform_service:
            api_key_value = self._platform_service._api_key
        if not api_key_value:
            raise AgentError(
                "API key is required for iOS app installation. "
                "Set MINITAP_API_KEY environment variable."
            )

        base_url = settings.MINITAP_BASE_URL or "https://platform.minitap.ai"

        app_name = app_path.stem
        asset_name = f"{app_name}.zip"

        logger.info(f"Preparing iOS app for upload: {app_name}")

        # Create a temporary zip file
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_path = Path(temp_dir) / asset_name
            zip_base = str(zip_path.with_suffix(""))
            shutil.make_archive(zip_base, "zip", app_path.parent, app_path.name)

            # Calculate MD5 hash
            md5 = hashlib.md5()
            with open(zip_path, "rb") as f:
                for chunk in iter(lambda: f.read(8192), b""):
                    md5.update(chunk)
            md5_hash = md5.hexdigest()

            logger.info(f"Created zip archive: {zip_path.name}, MD5: {md5_hash}")

            # Upload to Limrun using assets API
            client = AsyncLimrun(api_key=api_key_value, base_url=f"{base_url}/api/v1/limrun")
            try:
                logger.info("Getting upload URL from Limrun assets API...")
                asset_response = await client.assets.get_or_create(name=asset_name)

                # Check if we need to upload (MD5 mismatch or no existing file)
                existing_md5 = asset_response.md5
                if existing_md5 == md5_hash:
                    logger.info("App already uploaded with matching MD5, skipping upload")
                else:
                    upload_url = asset_response.signed_upload_url
                    if not upload_url:
                        raise AgentError("No upload URL returned from Limrun assets API")

                    logger.info("Uploading iOS app to Limrun storage...")
                    async with httpx.AsyncClient(timeout=300.0) as http_client:
                        file_content = zip_path.read_bytes()
                        response = await http_client.put(
                            upload_url,
                            content=file_content,
                            headers={"Content-Type": "application/zip"},
                        )
                        response.raise_for_status()
                    logger.info("iOS app uploaded successfully")

                download_url = asset_response.signed_download_url
                if not download_url:
                    raise AgentError("No download URL returned from Limrun assets API")

            finally:
                await client.close()

        # Install app via WebSocket
        logger.info("Installing app on Limrun iOS device...")
        install_result = await limrun_controller.client.install_app(url=download_url, md5=md5_hash)
        bundle_id = install_result.get("bundleId", "")

        if bundle_id:
            logger.success(f"iOS app installed successfully: {bundle_id}")
        else:
            logger.success("iOS app installed successfully")

        return bundle_id

    def new_task(self, goal: str):
        """
        Create a new task request builder.

        Args:
            goal: Natural language description of what to accomplish

        Returns:
            TaskRequestBuilder that can be configured with:
            - .with_output_format() for structured output
            - .with_output_description() for output description
            - .with_locked_app_package() to restrict execution to a specific app
            - .using_profile() to specify an LLM profile
            - .with_max_steps() to set maximum execution steps
            - .with_trace_recording() to enable trace recording
            - .with_name() to set a custom task name
        """
        return TaskRequestBuilder[None].from_common(
            goal=goal,
            common=self._config.task_request_defaults,
        )

    @overload
    async def run_task(
        self,
        *,
        goal: str,
        output: type[TOutput],
        profile: str | AgentProfile | None = None,
        name: str | None = None,
        locked_app_package: str | None = None,
        app_path: str | Path | None = None,
    ) -> TOutput | None: ...

    @overload
    async def run_task(
        self,
        *,
        goal: str,
        output: str,
        profile: str | AgentProfile | None = None,
        name: str | None = None,
        locked_app_package: str | None = None,
        app_path: str | Path | None = None,
    ) -> str | dict | None: ...

    @overload
    async def run_task(
        self,
        *,
        goal: str,
        output=None,
        profile: str | AgentProfile | None = None,
        name: str | None = None,
        locked_app_package: str | None = None,
        app_path: str | Path | None = None,
    ) -> str | None: ...

    @overload
    async def run_task(
        self,
        *,
        request: TaskRequest[None],
        locked_app_package: str | None = None,
        app_path: str | Path | None = None,
    ) -> str | dict | None: ...

    @overload
    async def run_task(
        self,
        *,
        request: TaskRequest[TOutput],
        locked_app_package: str | None = None,
        app_path: str | Path | None = None,
    ) -> TOutput | None: ...

    @overload
    async def run_task(
        self,
        *,
        request: PlatformTaskRequest[None],
        locked_app_package: str | None = None,
        app_path: str | Path | None = None,
    ) -> str | dict | None: ...

    @overload
    async def run_task(
        self,
        *,
        request: PlatformTaskRequest[TOutput],
        locked_app_package: str | None = None,
        app_path: str | Path | None = None,
    ) -> TOutput | None: ...

    async def run_task(
        self,
        *,
        goal: str | None = None,
        output: type[TOutput] | str | None = None,
        profile: str | AgentProfile | None = None,
        locked_app_package: str | None = None,
        name: str | None = None,
        app_path: str | Path | None = None,
        request: TaskRequest[TOutput] | PlatformTaskRequest[TOutput] | None = None,
    ) -> str | dict | TOutput | None:
        # Check if cloud mobile is configured
        if self._config.cloud_mobile_id_or_ref:
            if request is None or not isinstance(request, PlatformTaskRequest):
                raise AgentTaskRequestError(
                    "When using a cloud mobile, only PlatformTaskRequest is supported. "
                    "Use AgentConfigBuilder.for_cloud_mobile() only with PlatformTaskRequest."
                )
            # Use cloud mobile execution path
            return await self._run_cloud_mobile_task(
                request=request, locked_app_package=locked_app_package
            )

        # Normal local execution path
        if request is not None:
            task_info = None
            if isinstance(request, PlatformTaskRequest):
                if not self._platform_service:
                    raise PlatformServiceUninitializedError()
                task_info = await self._platform_service.create_task_run(
                    request=request,
                    locked_app_package=locked_app_package,
                    enable_video_tools=self._config.video_recording_enabled,
                )
                request.task_run_id = task_info.task_run.id
                request.task_run_id_available_event.set()
                self._config.agent_profiles[task_info.llm_profile.name] = task_info.llm_profile
                request = task_info.task_request
            elif locked_app_package is not None:
                if request.locked_app_package:
                    logger.warning(
                        "Locked app package specified both in the request and as a parameter. "
                        "Using the parameter value."
                    )
                request.locked_app_package = locked_app_package
            # Handle app_path parameter override
            if app_path is not None:
                if request.app_path:
                    logger.warning(
                        "App path specified both in the request and as a parameter. "
                        "Using the parameter value."
                    )
                request.app_path = Path(app_path) if isinstance(app_path, str) else app_path
            return await self._run_task(
                request=request, task_info=task_info, platform_service=self._platform_service
            )
        if goal is None:
            raise AgentTaskRequestError("Goal is required")
        task_request = self.new_task(goal=goal)
        if output is not None:
            if isinstance(output, str):
                task_request.with_output_description(description=output)
            elif output is not NoneType:
                task_request.with_output_format(output_format=output)
        if profile is not None:
            task_request.using_profile(profile=profile)
        if name is not None:
            task_request.with_name(name=name)
        if locked_app_package is not None:
            task_request.with_locked_app_package(package_name=locked_app_package)
        if app_path is not None:
            task_request.with_app_path(app_path=app_path)
        return await self._run_task(task_request.build())

    async def _run_cloud_mobile_task(
        self,
        request: PlatformTaskRequest[TOutput],
        locked_app_package: str | None = None,
    ) -> str | dict | TOutput | None:
        """
        Execute a task on a cloud mobile.

        This method triggers the task execution on the Platform and polls
        for completion without running any agentic logic locally.
        """
        if not self._cloud_mobile_id:
            raise AgentTaskRequestError("Cloud mobile ID is not configured")

        if not self._cloud_mobile_service:
            raise CloudMobileServiceUninitializedError()

        if not self._platform_service:
            raise PlatformServiceUninitializedError()

        if self._config.video_recording_enabled:
            profile_name = request.profile or DEFAULT_PROFILE_NAME
            _, profile = await self._platform_service.get_profile(profile_name)
            if not profile.llm_config.utils.video_analyzer:
                raise AgentTaskRequestError(
                    f"video_recording_enabled: profile '{profile_name}' "
                    "must have a video_analyzer agent configured"
                )

        # Start cloud mobile if not already started
        logger.info(f"Starting cloud mobile '{self._cloud_mobile_id}'...")
        await self._cloud_mobile_service.start_and_wait_for_ready(
            cloud_mobile_id=self._cloud_mobile_id,
        )
        logger.info(
            f"Starting cloud mobile task execution '{self._cloud_mobile_id}'",
        )

        def log_callback(message: str):
            """Callback for logging timeline updates."""
            logger.info(message)

        def status_callback(
            status: TaskRunStatus,
            status_message: str | None,
        ):
            """Callback for status updates."""
            logger.info(f"Task status update: [{status}] {status_message}")

        async def _execute_cloud(cloud_mobile_service: CloudMobileService, cloud_mobile_id: str):
            try:
                # Execute task on cloud mobile and wait for completion
                final_status, error, output = await cloud_mobile_service.run_task_on_cloud_mobile(
                    cloud_mobile_id=cloud_mobile_id,
                    request=request,
                    on_status_update=status_callback,
                    on_log=log_callback,
                    locked_app_package=locked_app_package,
                    enable_video_tools=self._config.video_recording_enabled,
                )
                if final_status == "completed":
                    logger.success("Cloud mobile task completed successfully")
                    return output
                if final_status == "failed":
                    logger.error(f"Cloud mobile task failed: {error}")
                    raise AgentTaskRequestError(
                        f"Task execution failed on cloud mobile: {error}",
                    )
                if final_status == "cancelled":
                    logger.warning("Cloud mobile task was cancelled")
                    raise AgentTaskRequestError("Task execution was cancelled")
                logger.error(f"Unknown cloud mobile task status: {final_status}")
                raise AgentTaskRequestError(f"Unknown task status: {final_status}")
            except asyncio.CancelledError:
                # Propagate cancellation to parent coroutine.
                logger.info("Task cancelled during execution, re-raising CancelledError")
                raise
            except AgentTaskRequestError as e:
                # Capture and re-raise known exceptions
                telemetry.capture_exception(e, {"phase": "cloud_mobile_task"})
                raise
            except Exception as e:
                logger.error(f"Unexpected error during cloud mobile task execution: {e}")
                telemetry.capture_exception(e, {"phase": "cloud_mobile_task"})
                raise AgentTaskRequestError(f"Unexpected error: {e}") from e

        async with self._task_lock:
            if self._current_task and not self._current_task.done():
                logger.warning(
                    "Another cloud task is running; cancelling it before starting new one.",
                )
                self.stop_current_task()
                try:
                    await self._current_task
                except asyncio.CancelledError:
                    pass
            try:
                self._current_task = asyncio.create_task(
                    _execute_cloud(
                        cloud_mobile_service=self._cloud_mobile_service,
                        cloud_mobile_id=self._cloud_mobile_id,
                    ),
                )
                return await self._current_task
            finally:
                self._current_task = None

    async def _run_task(
        self,
        request: TaskRequest[TOutput],
        task_info: PlatformTaskInfo | None = None,
        platform_service: PlatformService | None = None,
    ) -> str | dict | TOutput | None:
        if not self._initialized:
            raise AgentNotInitializedError()

        if request.profile:
            agent_profile = self._config.agent_profiles.get(request.profile)
            if agent_profile is None:
                raise AgentProfileNotFoundError(request.profile)
        else:
            agent_profile = self._config.default_profile

        if (
            self._config.video_recording_enabled
            and agent_profile.llm_config.utils.video_analyzer is None
        ):
            raise ValueError(
                f"with_video_recording_tools() requires 'video_analyzer' in utils for "
                f"profile '{agent_profile.name}'. Add 'video_analyzer' with a "
                f"video-capable model (e.g., gemini-3-flash-preview)."
            )

        logger.info(str(agent_profile))

        on_status_changed = None
        on_agent_thought = None
        on_plan_changes = None
        task_id = str(uuid.uuid4())
        if task_info:
            on_status_changed = self._get_task_status_change_callback(
                task_info=task_info, platform_service=platform_service
            )
            on_agent_thought = self._get_new_agent_thought_callback(
                task_info=task_info, platform_service=platform_service
            )
            on_plan_changes = self._get_plan_changes_callback(
                task_info=task_info, platform_service=platform_service
            )
            task_id = task_info.task_run.id

        task = Task(
            id=task_id,
            device=self._device_context,
            status="pending",
            request=request,
            created_at=datetime.now(),
            on_status_changed=on_status_changed,
        )
        self._tasks.append(task)
        task_name = task.get_name()

        # Extract API key from platform service if available
        api_key = None
        if platform_service:
            api_key = platform_service._api_key

        # Get limrun controller if available
        limrun_android_ctrl = None
        if hasattr(self, "_limrun_controller") and self._limrun_controller is not None:
            from minitap.mobile_use.controllers.limrun_controller import LimrunAndroidController

            if isinstance(self._limrun_controller, LimrunAndroidController):
                limrun_android_ctrl = self._limrun_controller

        context = MobileUseContext(
            trace_id=task.id,
            device=self._device_context,
            adb_client=self._adb_client,
            ui_adb_client=self._ui_adb_client,
            ios_client=self._ios_client,
            limrun_android_controller=limrun_android_ctrl,
            llm_config=agent_profile.llm_config,
            on_agent_thought=on_agent_thought,
            on_plan_changes=on_plan_changes,
            minitap_api_key=api_key,
            video_recording_enabled=(
                self._config.video_recording_enabled
                and agent_profile.llm_config.utils.video_analyzer is not None
            ),
        )

        self._prepare_tracing(task=task, context=context)
        await self._prepare_app_installation(task=task)
        await self._prepare_app_lock(task=task, context=context)
        self._prepare_output_files(task=task)

        output_config = None
        if request.output_description or request.output_format:
            output_config = OutputConfig(
                output_description=request.output_description,
                structured_output=request.output_format,  # type: ignore
            )
            logger.info(str(output_config))

        logger.info(f"[{task_name}] Starting graph with goal: `{request.goal}`")
        state = self._get_graph_state(task=task)
        graph_input = state.model_dump()
        task_start_time = datetime.now(UTC)

        telemetry.capture_task_started(
            task_id=task_id,
            platform=self._device_context.mobile_platform.value,
            has_locked_app=request.locked_app_package is not None,
        )

        async def _execute_task_logic():
            last_state: State | None = None
            last_state_snapshot: dict | None = None
            output = None
            try:
                logger.info(f"[{task_name}] Invoking graph with input: {graph_input}")
                await task.set_status(status="running", message="Invoking graph...")
                async for chunk in (await get_graph(context)).astream(
                    input=graph_input,
                    config={
                        "recursion_limit": task.request.max_steps,
                        "callbacks": self._config.graph_config_callbacks,
                    },
                    stream_mode=["messages", "custom", "updates", "values"],
                ):
                    stream_mode, payload = chunk
                    if stream_mode == "values":
                        last_state_snapshot = payload  # type: ignore
                        last_state = State(**last_state_snapshot)  # type: ignore
                        if task.request.thoughts_output_path:
                            record_events(
                                output_path=task.request.thoughts_output_path,
                                events=last_state.agents_thoughts,
                            )

                    if stream_mode == "updates":
                        for _, value in payload.items():  # type: ignore node name, node output
                            if value and "agents_thoughts" in value:
                                new_thoughts = value["agents_thoughts"]
                                last_item = new_thoughts[-1] if new_thoughts else None
                                if last_item:
                                    log_agent_thought(
                                        agent_thought=last_item,
                                    )

                if not last_state:
                    err = f"[{task_name}] No result received from graph"
                    logger.warning(err)
                    await task.finalize(content=output, state=last_state_snapshot, error=err)
                    return None

                print_ai_response_to_stderr(graph_result=last_state)
                output = await self._extract_output(
                    task_name=task_name,
                    ctx=context,
                    request=request,
                    output_config=output_config,
                    state=last_state,
                )
                logger.info(f"✅ Automation '{task_name}' is success ✅")
                await task.finalize(content=output, state=last_state_snapshot)
                duration = (datetime.now(UTC) - task_start_time).total_seconds()
                steps_count = len(last_state.agents_thoughts) if last_state else 0
                telemetry.capture_task_completed(
                    task_id=task_id,
                    success=True,
                    steps_count=steps_count,
                    duration_seconds=duration,
                )
                return output
            except asyncio.CancelledError:
                err = f"[{task_name}] Task cancelled"
                logger.warning(err)
                await task.finalize(
                    content=output,
                    state=last_state_snapshot,
                    error=err,
                    cancelled=True,
                )
                duration = (datetime.now(UTC) - task_start_time).total_seconds()
                steps_count = len(last_state.agents_thoughts) if last_state else 0
                telemetry.capture_task_completed(
                    task_id=task_id,
                    success=False,
                    steps_count=steps_count,
                    duration_seconds=duration,
                    cancelled=True,
                )
                raise
            except Exception as e:
                err = f"[{task_name}] Error running automation: {e}"
                logger.error(err)
                await task.finalize(
                    content=output,
                    state=last_state_snapshot,
                    error=err,
                )
                duration = (datetime.now(UTC) - task_start_time).total_seconds()
                steps_count = len(last_state.agents_thoughts) if last_state else 0
                telemetry.capture_task_completed(
                    task_id=task_id,
                    success=False,
                    steps_count=steps_count,
                    duration_seconds=duration,
                )
                telemetry.capture_exception(e, {"task_id": task_id})
                if telemetry._session_id:
                    logger.info(
                        "If you need support, please include this session ID: "
                        f"{telemetry._session_id}"
                    )
                raise
            finally:
                await self._finalize_tracing(task=task, context=context)

        async with self._task_lock:
            if self._current_task and not self._current_task.done():
                logger.warning(
                    "Another automation task is already running. "
                    "Stopping it before starting the new one."
                )
                self.stop_current_task()
                try:
                    await self._current_task
                except asyncio.CancelledError:
                    pass

            try:
                self._current_task = asyncio.create_task(_execute_task_logic())
                return await self._current_task
            finally:
                self._current_task = None

    def stop_current_task(self):
        """Requests cancellation of the currently running automation task."""
        if self._current_task and not self._current_task.done():
            logger.info("Requesting to stop the current automation task...")
            was_cancelled = self._current_task.cancel()
            if was_cancelled:
                logger.success("Cancellation request for the current task was sent.")
            else:
                logger.warning(
                    "Could not send cancellation request for the current task "
                    "(it may already be completing)."
                )
        else:
            logger.info("No active automation task to stop.")

    async def get_screenshot(self) -> Image.Image:
        """
        Capture a screenshot from the mobile device.

        For cloud mobiles, this method calls the mobile-manager endpoint.
        For local mobiles, it uses ADB (Android) or xcrun (iOS) directly.

        Returns:
            Screenshot as PIL Image

        Raises:
            AgentNotInitializedError: If the agent is not initialized
            PlatformServiceUninitializedError: If cloud mobile service is not available
            Exception: If screenshot capture fails
        """
        # Check if cloud mobile is configured
        if self._cloud_mobile_id:
            if not self._cloud_mobile_service:
                raise CloudMobileServiceUninitializedError()
            screenshot = await self._cloud_mobile_service.get_screenshot(
                cloud_mobile_id=self._cloud_mobile_id,
            )
            return screenshot

        # Local device - use ADB or xcrun directly
        if not self._initialized:
            raise AgentNotInitializedError()

        if self._device_context.mobile_platform == DevicePlatform.ANDROID:
            # Use ADB to capture screenshot
            logger.info("Capturing screenshot from local Android device")
            if not self._adb_client:
                raise Exception("ADB client not initialized")

            device = self._adb_client.device(serial=self._device_context.device_id)
            screenshot = await asyncio.to_thread(device.screenshot)
            logger.info("Screenshot captured from local Android device")
            return screenshot

        elif self._device_context.mobile_platform == DevicePlatform.IOS:
            from io import BytesIO

            from minitap.mobile_use.controllers.limrun_controller import LimrunIosController

            # Check if using Limrun iOS controller
            if isinstance(self._ios_client, LimrunIosController):
                logger.info("Capturing screenshot from Limrun iOS device")
                screenshot_bytes = await self._ios_client.screenshot()
                if screenshot_bytes is None:
                    raise Exception("Failed to capture screenshot from Limrun iOS device")
                screenshot = Image.open(BytesIO(screenshot_bytes))
                logger.info("Screenshot captured from Limrun iOS device")
                return screenshot

            # Use xcrun to capture screenshot for local simulators
            import functools
            import subprocess

            logger.info("Capturing screenshot from local iOS device")
            try:
                # xcrun simctl io <device> screenshot --type=png -
                result = await asyncio.to_thread(
                    functools.partial(
                        subprocess.run,
                        [
                            "xcrun",
                            "simctl",
                            "io",
                            self._device_context.device_id,
                            "screenshot",
                            "--type=png",
                            "-",
                        ],
                        capture_output=True,
                        check=True,
                    )
                )
                # Convert bytes to PIL Image
                screenshot = Image.open(BytesIO(result.stdout))
                logger.info("Screenshot captured from local iOS device")
                return screenshot
            except subprocess.CalledProcessError as e:
                logger.error(f"Failed to capture screenshot: {e}")
                raise Exception(f"Failed to capture screenshot from iOS device: {e}")

        else:
            raise Exception(f"Unsupported platform: {self._device_context.mobile_platform}")

    async def clean(self, force: bool = False):
        if self._cloud_mobile_id:
            self._initialized = False
            logger.info("✅ Cloud-mode agent stopped.")
            # End telemetry session if started by SDK (not CLI)
            if telemetry._session_id and telemetry._session_context.get("source") == "sdk":
                telemetry.end_session(success=True)
            return

        # Cleanup Limrun device if provisioned by SDK
        if self._limrun_instance_id:
            await self._cleanup_limrun_device()
            self._initialized = False
            logger.info("✅ Limrun agent stopped.")
            if telemetry._session_id and telemetry._session_context.get("source") == "sdk":
                telemetry.end_session(success=True)
            return

        if not self._initialized and not force:
            return

        if self._ios_client:
            await self._ios_client.cleanup()
            self._ios_client = None

        self._initialized = False
        logger.info("✅ Mobile-use agent stopped.")

        # End telemetry session if started by SDK (not CLI)
        if telemetry._session_id and telemetry._session_context.get("source") == "sdk":
            telemetry.end_session(success=True)

    async def _init_limrun_device(
        self,
        android_controller: LimrunAndroidController | None = None,
        ios_controller: LimrunIosController | None = None,
    ) -> bool:
        """
        Initialize a Limrun cloud device.

        This method either uses pre-configured controllers or provisions a new
        Limrun device based on limrun_config, connects to it, and sets up the
        appropriate clients for device interaction.

        Args:
            android_controller: Pre-configured Limrun Android controller (optional).
            ios_controller: Pre-configured Limrun iOS controller (optional).
        """

        # Use pre-configured Android controller
        if android_controller is not None:
            logger.info("Using pre-configured Limrun Android controller")
            self._limrun_controller = android_controller
            self._limrun_instance_id = android_controller.instance_id

            self._adb_client = android_controller._adb_client
            self._ui_adb_client = android_controller._ui_client
            self._ios_client = None
            self._ios_device_type = None

            self._device_context = DeviceContext(
                host_platform="LINUX",
                mobile_platform=DevicePlatform.ANDROID,
                device_id=android_controller._adb_serial or self._limrun_instance_id,
                device_width=android_controller.device_width,
                device_height=android_controller.device_height,
            )

            logger.info(f"Limrun Android device ready: {self._limrun_instance_id}")
            logger.info(self._device_context.to_str())
            logger.info("✅ Mobile-use agent initialized with pre-configured Limrun Android.")
            self._initialized = True
            telemetry.capture_agent_initialized(
                platform=DevicePlatform.ANDROID.value,
                device_id=self._limrun_instance_id,
            )
            return True

        # Use pre-configured iOS controller
        if ios_controller is not None:
            logger.info("Using pre-configured Limrun iOS controller")
            self._limrun_controller = ios_controller
            self._limrun_instance_id = ios_controller.instance_id

            self._adb_client = None
            self._ui_adb_client = None
            self._ios_client = ios_controller
            self._ios_device_type = DeviceType.LIMRUN

            self._device_context = DeviceContext(
                host_platform="LINUX",
                mobile_platform=DevicePlatform.IOS,
                device_id=self._limrun_instance_id,
                device_width=ios_controller.device_width,
                device_height=ios_controller.device_height,
            )

            logger.info(f"Limrun iOS device ready: {self._limrun_instance_id}")
            logger.info(self._device_context.to_str())
            logger.info("✅ Mobile-use agent initialized with pre-configured Limrun iOS.")
            self._initialized = True
            telemetry.capture_agent_initialized(
                platform=DevicePlatform.IOS.value,
                device_id=self._limrun_instance_id,
            )
            return True

        # Fall back to provisioning via limrun_config
        from minitap.mobile_use.clients.limrun_factory import (
            LimrunInstanceConfig,
            create_limrun_android_instance,
            create_limrun_ios_instance,
        )

        limrun_config = self._config.limrun_config
        if limrun_config is None:
            raise ValueError("limrun_config is not set and no pre-configured controller provided")

        logger.info(f"Provisioning Limrun {limrun_config.platform.value} device...")

        instance_config = LimrunInstanceConfig(
            api_key=limrun_config.api_key,
            base_url=limrun_config.base_url,
            inactivity_timeout=limrun_config.inactivity_timeout,
            hard_timeout=limrun_config.hard_timeout,
            display_name=limrun_config.display_name,
            labels=limrun_config.labels,
        )

        if limrun_config.platform == LimrunPlatform.ANDROID:
            instance, controller = await create_limrun_android_instance(instance_config)
            self._limrun_instance_id = instance.metadata.id
            self._limrun_controller = controller
            try:
                await controller.connect()
            except Exception:
                await self._cleanup_limrun_device()
                raise

            self._adb_client = controller._adb_client
            self._ui_adb_client = controller._ui_client
            self._ios_client = None
            self._ios_device_type = None

            self._device_context = DeviceContext(
                host_platform="LINUX",
                mobile_platform=DevicePlatform.ANDROID,
                device_id=controller._adb_serial or self._limrun_instance_id,
                device_width=controller.device_width,
                device_height=controller.device_height,
            )
        else:
            instance, controller, limrun_ctrl = await create_limrun_ios_instance(instance_config)
            self._limrun_instance_id = instance.metadata.id
            self._limrun_controller = limrun_ctrl  # Store underlying controller for cleanup

            self._adb_client = None
            self._ui_adb_client = None
            self._ios_client = controller.ios_client  # Use the LimrunIosController as ios_client
            self._ios_device_type = DeviceType.LIMRUN

            self._device_context = DeviceContext(
                host_platform="LINUX",
                mobile_platform=DevicePlatform.IOS,
                device_id=self._limrun_instance_id,
                device_width=controller.device_width,
                device_height=controller.device_height,
            )

        logger.info(
            f"Limrun {limrun_config.platform.value} device ready: {self._limrun_instance_id}"
        )
        logger.info(self._device_context.to_str())
        logger.info("✅ Mobile-use agent initialized with Limrun.")
        self._initialized = True
        telemetry.capture_agent_initialized(
            platform=limrun_config.platform.value,
            device_id=self._limrun_instance_id,
        )
        return True

    async def _cleanup_limrun_device(self) -> None:
        """Cleanup Limrun device resources."""
        from minitap.mobile_use.clients.limrun_factory import (
            LimrunInstanceConfig,
            delete_limrun_android_instance,
            delete_limrun_ios_instance,
        )

        # Always cleanup controller if present
        if self._limrun_controller:
            logger.info("Cleaning up Limrun controller...")
            await self._limrun_controller.cleanup()
            self._limrun_controller = None

        # Only attempt instance deletion if limrun_config is present
        if self._config.limrun_config and self._limrun_instance_id:
            logger.info(f"Deleting Limrun instance: {self._limrun_instance_id}")
            limrun_config = self._config.limrun_config
            instance_config = LimrunInstanceConfig(
                api_key=limrun_config.api_key,
                base_url=limrun_config.base_url,
            )

            try:
                if limrun_config.platform == LimrunPlatform.ANDROID:
                    await delete_limrun_android_instance(instance_config, self._limrun_instance_id)
                else:
                    await delete_limrun_ios_instance(instance_config, self._limrun_instance_id)
            except Exception as e:
                logger.warning(f"Failed to delete Limrun instance: {e}")
        elif self._limrun_instance_id:
            logger.info(
                f"Skipping Limrun instance deletion (no limrun_config): {self._limrun_instance_id}"
            )

        self._limrun_instance_id = None

    async def _prepare_app_installation(self, task: Task) -> str | None:
        """Install app if app_path is specified in the task request.

        Returns:
            The bundle ID of the installed app (iOS only), or None.
        """
        if not task.request.app_path:
            return None

        task_name = task.get_name()
        logger.info(f"[{task_name}] Installing app from: {task.request.app_path}")

        bundle_id = await self.install_app(task.request.app_path)

        if bundle_id:
            logger.info(f"[{task_name}] App installed with bundle ID: {bundle_id}")
            # If locked_app_package is not set, automatically lock to the installed app
            if not task.request.locked_app_package:
                logger.info(f"[{task_name}] Auto-locking to installed app: {bundle_id}")
                task.request.locked_app_package = bundle_id

        return bundle_id

    async def _prepare_app_lock(self, task: Task, context: MobileUseContext):
        """Prepare app lock by launching the locked app if specified."""
        if not task.request.locked_app_package:
            return

        task_name = task.get_name()
        logger.info(f"[{task_name}] Preparing app lock for: {task.request.locked_app_package}")

        app_lock_status = await _handle_initial_app_launch(
            ctx=context, locked_app_package=task.request.locked_app_package
        )

        if context.execution_setup is None:
            context.execution_setup = ExecutionSetup(app_lock_status=app_lock_status)
        else:
            context.execution_setup.app_lock_status = app_lock_status

        if app_lock_status.locked_app_initial_launch_success is False:
            error = app_lock_status.locked_app_initial_launch_error
            logger.warning(f"[{task_name}] Failed to launch locked app: {error}")

    def _prepare_tracing(self, task: Task, context: MobileUseContext):
        """Prepare tracing setup if record_trace is enabled."""
        if not task.request.record_trace:
            return

        task_name = task.get_name()
        temp_trace_path = Path(self._tmp_traces_dir / task_name).resolve()
        traces_output_path = Path(task.request.trace_path).resolve()
        logger.info(f"[{task_name}] 📂 Traces output path: {traces_output_path}")
        logger.info(f"[{task_name}] 📄📂 Traces temp path: {temp_trace_path}")
        traces_output_path.mkdir(parents=True, exist_ok=True)
        temp_trace_path.mkdir(parents=True, exist_ok=True)

        context.execution_setup = ExecutionSetup(
            traces_path=self._tmp_traces_dir,
            trace_name=task_name,
            enable_remote_tracing=task.request.enable_remote_tracing,
        )

    async def _finalize_tracing(self, task: Task, context: MobileUseContext):
        exec_setup_ctx = context.execution_setup
        if not exec_setup_ctx:
            return

        if exec_setup_ctx.traces_path is None or exec_setup_ctx.trace_name is None:
            return

        task_name = task.get_name()
        status = "_PASS" if task.status == "completed" else "_FAIL"
        ts = task.created_at.strftime("%Y-%m-%dT%H-%M-%S")
        new_name = f"{exec_setup_ctx.trace_name}{status}_{ts}"

        temp_trace_path = (self._tmp_traces_dir / exec_setup_ctx.trace_name).resolve()
        traces_output_path = Path(task.request.trace_path).resolve()

        logger.info(f"[{task_name}] Compiling trace FROM FOLDER: " + str(temp_trace_path))
        create_gif_from_trace_folder(temp_trace_path)
        create_steps_json_from_trace_folder(temp_trace_path)

        if exec_setup_ctx.enable_remote_tracing:
            gif_path = temp_trace_path / "trace.gif"
            if gif_path.exists() and self._platform_service:
                try:
                    task_run_id = await self._platform_service.upload_trace_gif(
                        task_run_id=task.id, gif_path=gif_path
                    )
                    if task_run_id:
                        platform_url = f"{settings.MINITAP_BASE_URL}/task-runs/{task_run_id}"
                        logger.info(f"[{task_name}] 🌐 View on platform: {platform_url}")
                except Exception as e:
                    logger.warning(f"[{task_name}] Failed to upload trace GIF: {e}")

        logger.info(f"[{task_name}] Video created, removing dust...")
        remove_images_from_trace_folder(temp_trace_path)
        remove_steps_json_from_trace_folder(temp_trace_path)
        logger.info(f"[{task_name}] 📽️ Trace compiled, moving to output path 📽️")

        output_folder_path = temp_trace_path.rename(traces_output_path / new_name).resolve()
        logger.info(f"[{task_name}] 📂✅ Traces located in: {output_folder_path}")

    def _prepare_output_files(self, task: Task):
        if task.request.llm_output_path:
            _validate_and_prepare_file(file_path=task.request.llm_output_path)
        if task.request.thoughts_output_path:
            _validate_and_prepare_file(file_path=task.request.thoughts_output_path)

    async def _extract_output(
        self,
        task_name: str,
        ctx: MobileUseContext,
        request: TaskRequest[TOutput],
        output_config: OutputConfig | None,
        state: State,
    ) -> str | dict | TOutput | None:
        if output_config and output_config.needs_structured_format():
            logger.info(f"[{task_name}] Generating structured output...")
            try:
                structured_output = await outputter(
                    ctx=ctx,
                    output_config=output_config,
                    graph_output=state,
                )
                logger.info(f"[{task_name}] Structured output: {structured_output}")
                record_events(output_path=request.llm_output_path, events=structured_output)
                if request.output_format is not None and request.output_format is not NoneType:
                    return request.output_format.model_validate(structured_output)
                return structured_output
            except Exception as e:
                logger.error(f"[{task_name}] Failed to generate structured output: {e}")
                return None
        if state and state.agents_thoughts:
            last_msg = state.agents_thoughts[-1]
            logger.info(str(last_msg))
            record_events(output_path=request.llm_output_path, events=last_msg)
            return last_msg
        return None

    def _get_graph_state(self, task: Task):
        return State(
            messages=[],
            initial_goal=task.request.goal,
            subgoal_plan=[],
            latest_ui_hierarchy=None,
            latest_screenshot=None,
            focused_app_info=None,
            device_date=None,
            structured_decisions=None,
            complete_subgoals_by_ids=[],
            agents_thoughts=[],
            remaining_steps=task.request.max_steps,
            executor_messages=[],
            cortex_last_thought=None,
            scratchpad={},
        )

    def _init_clients(
        self,
        device_id: str,
        platform: DevicePlatform,
        ios_device_type: DeviceType | None,
        retry_count: int,
        retry_wait_seconds: int,
    ):
        self._adb_client = (
            AdbClient(host=self._config.servers.adb_host, port=self._config.servers.adb_port)
            if platform == DevicePlatform.ANDROID
            else None
        )
        self._ui_adb_client = (
            UIAutomatorClient(device_id=device_id) if platform == DevicePlatform.ANDROID else None
        )

        # Initialize iOS client using factory (auto-detects device type if not provided)
        if platform == DevicePlatform.IOS:
            self._ios_client = get_ios_client(
                udid=device_id,
                config=self._config.ios_client_config,
            )
            self._ios_device_type = ios_device_type or (
                DeviceType.PHYSICAL
                if isinstance(self._ios_client, WdaClientWrapper)
                else DeviceType.SIMULATOR
            )
        else:
            self._ios_client = None
            self._ios_device_type = None

    def _run_servers(self, device_id: str, platform: DevicePlatform) -> bool:
        if platform == DevicePlatform.ANDROID:
            if not self._ui_adb_client:
                error_msg = (
                    "UIAutomator client is required for Android but not available. "
                    "Please ensure UIAutomator2 is properly installed and configured."
                )
                logger.error(error_msg)
                raise ValueError(error_msg)
            logger.info("✓ UIAutomator client available for Android")
        elif platform == DevicePlatform.IOS:
            if not self._ios_client:
                error_msg = (
                    "iOS client is required but not available. "
                    "Ensure idb (simulators) or WDA (physical) is available."
                )
                logger.error(error_msg)
                raise ValueError(error_msg)
            if isinstance(self._ios_client, WdaClientWrapper):
                client_type = "WDA"
            elif isinstance(self._ios_client, BrowserStackClientWrapper):
                client_type = "BrowserStack"
            else:
                client_type = "IDB"
            logger.info(f"✓ iOS client available ({client_type})")

        return True

    async def _get_device_context(
        self,
        device_id: str,
        platform: DevicePlatform,
    ) -> DeviceContext:
        from platform import system

        host_platform = system()

        # Get real device dimensions from the device
        if platform == DevicePlatform.ANDROID:
            if self._ui_adb_client:
                try:
                    # Use UIAutomator to get actual screen dimensions
                    screen_data = self._ui_adb_client.get_screen_data()
                    device_width = screen_data.width
                    device_height = screen_data.height
                    logger.info(
                        f"Retrieved Android screen dimensions: {device_width}x{device_height}"
                    )
                except Exception as e:
                    logger.warning(f"Failed to get Android screen dimensions: {e}, using defaults")
                    device_width, device_height = 1080, 2340
            else:
                logger.warning("UIAutomator client not available, using default dimensions")
                device_width, device_height = 1080, 2340
        else:  # iOS
            if self._ios_client:
                try:
                    # Use iOS client to take a screenshot and get dimensions
                    screenshot_data = await self._ios_client.screenshot()  # type: ignore[call-arg]
                    if screenshot_data:
                        img = Image.open(BytesIO(screenshot_data))
                        device_width = img.width
                        device_height = img.height
                        logger.info(
                            f"Retrieved iOS screen dimensions: {device_width}x{device_height}"
                        )
                    else:
                        logger.warning("IDB screenshot returned None, using default dimensions")
                        device_width, device_height = 375, 812
                except Exception as e:
                    logger.warning(f"Failed to get iOS screen dimensions: {e}, using defaults")
                    device_width, device_height = 375, 812
            else:
                logger.warning("IDB client not available, using default dimensions")
                device_width, device_height = 375, 812

        return DeviceContext(
            host_platform="WINDOWS" if host_platform == "Windows" else "LINUX",
            mobile_platform=platform,
            device_id=device_id,
            device_width=device_width,
            device_height=device_height,
        )

    def _get_task_status_change_callback(
        self,
        task_info: PlatformTaskInfo,
        platform_service: PlatformService | None = None,
    ) -> Callable[[TaskRunStatus, str | None, Any | None], Coroutine]:
        service = platform_service or self._platform_service

        async def change_status(
            status: TaskRunStatus,
            message: str | None = None,
            output: Any | None = None,
        ):
            if not service:
                raise PlatformServiceUninitializedError()
            try:
                await service.update_task_run_status(
                    task_run_id=task_info.task_run.id,
                    status=status,
                    message=message,
                    output=output,
                )
            except Exception as e:
                logger.error(f"Failed to update task run status: {e}")

        return change_status

    def _get_plan_changes_callback(
        self,
        task_info: PlatformTaskInfo,
        platform_service: PlatformService | None = None,
    ) -> Callable[[list[Subgoal], IsReplan], Coroutine]:
        service = platform_service or self._platform_service
        current_plan: TaskRunPlanResponse | None = None

        async def update_plan(plan: list[Subgoal], is_replan: IsReplan):
            nonlocal current_plan

            if not service:
                raise PlatformServiceUninitializedError()
            try:
                if is_replan and current_plan:
                    # End previous plan
                    await service.upsert_task_run_plan(
                        task_run_id=task_info.task_run.id,
                        started_at=current_plan.started_at,
                        plan=plan,
                        ended_at=datetime.now(UTC),
                        plan_id=current_plan.id,
                    )
                    current_plan = None

                current_plan = await service.upsert_task_run_plan(
                    task_run_id=task_info.task_run.id,
                    started_at=current_plan.started_at if current_plan else datetime.now(UTC),
                    plan=plan,
                    ended_at=current_plan.ended_at if current_plan else None,
                    plan_id=current_plan.id if current_plan else None,
                )
            except Exception as e:
                logger.error(f"Failed to update plan: {e}")

        return update_plan

    def _get_new_agent_thought_callback(
        self,
        task_info: PlatformTaskInfo,
        platform_service: PlatformService | None = None,
    ) -> Callable[[AgentNode, str], Coroutine]:
        service = platform_service or self._platform_service

        async def add_agent_thought(agent: AgentNode, thought: str):
            if not service:
                raise PlatformServiceUninitializedError()
            try:
                await service.add_agent_thought(
                    task_run_id=task_info.task_run.id,
                    agent=agent,
                    thought=thought,
                )
            except Exception as e:
                logger.error(f"Failed to add agent thought: {e}")

        return add_agent_thought


def _validate_and_prepare_file(file_path: Path):
    path_obj = Path(file_path)
    if path_obj.exists() and path_obj.is_dir():
        raise AgentTaskRequestError(f"Error: Path '{file_path}' is a directory, not a file.")
    try:
        path_obj.parent.mkdir(parents=True, exist_ok=True)
        path_obj.touch(exist_ok=True)
    except OSError as e:
        raise AgentTaskRequestError(f"Error creating file '{file_path}': {e}")


def print_ai_response_to_stderr(graph_result: State):
    for msg in reversed(graph_result.messages):
        if isinstance(msg, AIMessage):
            print(msg.content, file=sys.stderr)
            return



import asyncio
import os
from enum import StrEnum
from shutil import which
from typing import Annotated

import typer
from adbutils import AdbClient
from langchain_core.callbacks.base import Callbacks
from rich.console import Console

from minitap.mobile_use.clients.ios_client_config import (
    IdbClientConfig,
    IosClientConfig,
    WdaClientConfig,
)
from minitap.mobile_use.clients.limrun_factory import (
    LimrunInstanceConfig,
    LimrunPlatform,
    create_limrun_android_instance,
    create_limrun_ios_instance,
    delete_limrun_android_instance,
    delete_limrun_ios_instance,
)
from minitap.mobile_use.config import initialize_llm_config, settings
from minitap.mobile_use.sdk import Agent
from minitap.mobile_use.sdk.builders import Builders
from minitap.mobile_use.sdk.types.task import AgentProfile
from minitap.mobile_use.services.telemetry import telemetry
from minitap.mobile_use.utils.cli_helpers import display_device_status
from minitap.mobile_use.utils.logger import get_logger
from minitap.mobile_use.utils.video import check_ffmpeg_available

app = typer.Typer(add_completion=False, pretty_exceptions_enable=False)
logger = get_logger(__name__)


class DeviceType(StrEnum):
    """Device type for mobile-use agent."""

    LOCAL = "local"
    LIMRUN = "limrun"


async def run_automation(
    goal: str,
    locked_app_package: str | None = None,
    test_name: str | None = None,
    traces_output_path_str: str = "traces",
    output_description: str | None = None,
    graph_config_callbacks: Callbacks = [],
    video_recording_tools_enabled: bool = False,
    wda_url: str | None = None,
    wda_timeout: float | None = None,
    wda_auto_start_iproxy: bool | None = None,
    wda_auto_start_wda: bool | None = None,
    wda_project_path: str | None = None,
    wda_startup_timeout: float | None = None,
    idb_host: str | None = None,
    idb_port: int | None = None,
    device_type: DeviceType = DeviceType.LOCAL,
    limrun_platform: LimrunPlatform | None = None,
):
    llm_config = initialize_llm_config()
    agent_profile = AgentProfile(name="default", llm_config=llm_config)
    config = Builders.AgentConfig.with_default_profile(profile=agent_profile)
    if video_recording_tools_enabled:
        config.with_video_recording_tools()

    # Limrun device provisioning
    limrun_instance_id: str | None = None
    limrun_controller = None
    limrun_config: LimrunInstanceConfig | None = None

    if device_type == DeviceType.LIMRUN:
        if limrun_platform is None:
            raise ValueError("--limrun-platform is required when using --device-type limrun")

        logger.info(f"Provisioning Limrun {limrun_platform.value} device...")
        limrun_config = LimrunInstanceConfig()

        if limrun_platform == LimrunPlatform.ANDROID:
            instance, limrun_controller = await create_limrun_android_instance(limrun_config)
            limrun_instance_id = instance.metadata.id
            await limrun_controller.connect()
            config.with_limrun_android_controller(limrun_controller)
        else:
            instance, _, limrun_controller = await create_limrun_ios_instance(limrun_config)
            limrun_instance_id = instance.metadata.id
            # Connection is done in the factory, no need to call connect()
            config.with_limrun_ios_controller(limrun_controller)

        logger.info(f"Limrun {limrun_platform.value} device ready: {limrun_instance_id}")
    else:
        # Build iOS client config from CLI options (local device)
        wda_config = WdaClientConfig.with_overrides(
            wda_url=wda_url,
            timeout=wda_timeout,
            auto_start_iproxy=wda_auto_start_iproxy,
            auto_start_wda=wda_auto_start_wda,
            wda_project_path=wda_project_path,
            wda_startup_timeout=wda_startup_timeout,
        )
        idb_config = IdbClientConfig.with_overrides(host=idb_host, port=idb_port)
        config.with_ios_client_config(IosClientConfig(wda=wda_config, idb=idb_config))

        if settings.ADB_HOST:
            config.with_adb_server(host=settings.ADB_HOST, port=settings.ADB_PORT)

    if graph_config_callbacks:
        config.with_graph_config_callbacks(graph_config_callbacks)

    agent: Agent | None = None
    try:
        agent = Agent(config=config.build())
        await agent.init(
            retry_count=int(os.getenv("MOBILE_USE_HEALTH_RETRIES", 5)),
            retry_wait_seconds=int(os.getenv("MOBILE_USE_HEALTH_DELAY", 2)),
        )

        task = agent.new_task(goal)
        if locked_app_package:
            task.with_locked_app_package(locked_app_package)
        if test_name:
            task.with_name(test_name).with_trace_recording(path=traces_output_path_str)
        if output_description:
            task.with_output_description(output_description)

        agent_thoughts_path = os.getenv("EVENTS_OUTPUT_PATH", None)
        llm_result_path = os.getenv("RESULTS_OUTPUT_PATH", None)
        if agent_thoughts_path:
            task.with_thoughts_output_saving(path=agent_thoughts_path)
        if llm_result_path:
            task.with_llm_output_saving(path=llm_result_path)

        await agent.run_task(request=task.build())
    finally:
        if agent is not None:
            await agent.clean()

        # Cleanup Limrun device
        if limrun_instance_id and limrun_config:
            logger.info(f"Cleaning up Limrun device: {limrun_instance_id}")
            if limrun_controller:
                await limrun_controller.cleanup()
            if limrun_platform == LimrunPlatform.ANDROID:
                await delete_limrun_android_instance(limrun_config, limrun_instance_id)
            else:
                await delete_limrun_ios_instance(limrun_config, limrun_instance_id)


@app.command()
def main(
    goal: Annotated[str, typer.Argument(help="The main goal for the agent to achieve.")],
    test_name: Annotated[
        str | None,
        typer.Option(
            "--test-name",
            "-n",
            help="A name for the test recording. If provided, a trace will be saved.",
        ),
    ] = None,
    traces_path: Annotated[
        str,
        typer.Option(
            "--traces-path",
            "-p",
            help="The path to save the traces.",
        ),
    ] = "traces",
    output_description: Annotated[
        str | None,
        typer.Option(
            "--output-description",
            "-o",
            help=(
                """
                A dict output description for the agent.
                Ex: a JSON schema with 2 keys: type, price
                """
            ),
        ),
    ] = None,
    wda_url: Annotated[
        str | None,
        typer.Option(
            "--wda-url",
            help="Override WebDriverAgent URL (e.g. http://localhost:8100).",
        ),
    ] = None,
    wda_timeout: Annotated[
        float | None,
        typer.Option(
            "--wda-timeout",
            help="Timeout (seconds) for WDA operations.",
        ),
    ] = None,
    wda_auto_start_iproxy: Annotated[
        bool | None,
        typer.Option(
            "--wda-auto-start-iproxy/--no-wda-auto-start-iproxy",
            help="Auto-start iproxy if not running.",
        ),
    ] = None,
    wda_auto_start_wda: Annotated[
        bool | None,
        typer.Option(
            "--wda-auto-start-wda/--no-wda-auto-start-wda",
            help="Auto-build and run WDA via xcodebuild if not responding.",
        ),
    ] = None,
    wda_project_path: Annotated[
        str | None,
        typer.Option(
            "--wda-project-path",
            help="Path to WebDriverAgent.xcodeproj.",
        ),
    ] = None,
    wda_startup_timeout: Annotated[
        float | None,
        typer.Option(
            "--wda-startup-timeout",
            help="Timeout (seconds) while waiting for WDA to start.",
        ),
    ] = None,
    idb_host: Annotated[
        str | None,
        typer.Option(
            "--idb-host",
            help="IDB companion host (for simulators).",
        ),
    ] = None,
    idb_port: Annotated[
        int | None,
        typer.Option(
            "--idb-port",
            help="IDB companion port (for simulators).",
        ),
    ] = None,
    with_video_recording_tools: Annotated[
        bool,
        typer.Option(
            "--with-video-recording-tools",
            help="Enable AI agents to use video recording tools"
            " to analyze dynamic content on the screen.",
        ),
    ] = False,
    device_type: Annotated[
        DeviceType,
        typer.Option(
            "--device-type",
            "-d",
            help="Device type: 'local' for connected devices, 'limrun' for cloud devices.",
        ),
    ] = DeviceType.LOCAL,
    limrun_platform: Annotated[
        LimrunPlatform | None,
        typer.Option(
            "--limrun-platform",
            help="Platform for Limrun cloud device: 'android' or 'ios'. "
            "Required when --device-type is 'limrun'.",
        ),
    ] = None,
):
    """
    Run the Mobile-use agent to automate tasks on a mobile device.
    """
    if with_video_recording_tools:
        check_ffmpeg_available()

    console = Console()

    if device_type == DeviceType.LOCAL:
        adb_client = None
        try:
            if which("adb"):
                adb_client = AdbClient(
                    host=settings.ADB_HOST or "localhost",
                    port=settings.ADB_PORT or 5037,
                )
        except Exception:
            pass  # ADB not available, will only support iOS devices

        display_device_status(console, adb_client=adb_client)
    else:
        if limrun_platform is None:
            console.print(
                "[red]Error: --limrun-platform is required when using --device-type limrun[/red]"
            )
            raise typer.Exit(1)
        console.print(f"[cyan]Using Limrun cloud device ({limrun_platform.value})...[/cyan]")

    # Start telemetry session with CLI context (only non-sensitive flags)
    session_id = telemetry.start_session(
        {
            "source": "cli",
            "has_output_description": output_description is not None,
        }
    )

    error_message = None
    cancelled = False
    try:
        asyncio.run(
            run_automation(
                goal=goal,
                test_name=test_name,
                traces_output_path_str=traces_path,
                output_description=output_description,
                wda_url=wda_url,
                wda_timeout=wda_timeout,
                wda_auto_start_iproxy=wda_auto_start_iproxy,
                wda_auto_start_wda=wda_auto_start_wda,
                wda_project_path=wda_project_path,
                wda_startup_timeout=wda_startup_timeout,
                idb_host=idb_host,
                idb_port=idb_port,
                video_recording_tools_enabled=with_video_recording_tools,
                device_type=device_type,
                limrun_platform=limrun_platform,
            )
        )
    except KeyboardInterrupt:
        cancelled = True
        error_message = "Task cancelled by user"
    except Exception as e:
        error_message = str(e)
        console.print(
            f"\n[dim]If you need support, please include this session ID: {session_id}[/dim]"
        )
        raise
    finally:
        telemetry.end_session(
            success=error_message is None,
            error=error_message,
        )
        if cancelled:
            raise SystemExit(130)


def _prompt_telemetry_consent(console: Console) -> None:
    """Prompt user for telemetry consent if not yet configured."""
    if not telemetry.needs_consent:
        return

    console.print()
    console.print("[bold]📊 Help improve mobile-use[/bold]")
    console.print(
        "We collect anonymous usage data to help debug and improve the SDK.\n"
        "No personal data, prompts, or device content is collected.\n"
        "You can change this anytime by setting MOBILE_USE_TELEMETRY_ENABLED=false\n"
    )

    try:
        import inquirer

        questions = [
            inquirer.Confirm(
                "consent",
                message="Enable anonymous telemetry?",
                default=True,
            )
        ]
        answers = inquirer.prompt(questions)
        if answers is not None:
            enabled = answers.get("consent", False)
            telemetry.set_consent(enabled)
            if enabled:
                console.print("[green]✓ Telemetry enabled. Thank you![/green]\n")
            else:
                console.print("[dim]Telemetry disabled.[/dim]\n")
        else:
            telemetry.set_consent(False)
    except (ImportError, KeyboardInterrupt):
        telemetry.set_consent(False)


def cli():
    console = Console()
    _prompt_telemetry_consent(console)
    telemetry.initialize()
    try:
        app()
    finally:
        telemetry.shutdown()


if __name__ == "__main__":
    cli()



import asyncio
import base64
import re
import tempfile
import time
from io import BytesIO
from pathlib import Path

from adbutils import AdbClient, AdbDevice
from PIL import Image

from minitap.mobile_use.clients.ui_automator_client import UIAutomatorClient
from minitap.mobile_use.controllers.device_controller import (
    MobileDeviceController,
    ScreenDataResponse,
)
from minitap.mobile_use.controllers.types import Bounds, CoordinatesSelectorRequest, TapOutput
from minitap.mobile_use.utils.logger import get_logger
from minitap.mobile_use.utils.video import (
    ANDROID_MAX_RECORDING_DURATION_SECONDS,
    DEFAULT_MAX_DURATION_SECONDS,
    VIDEO_READY_DELAY_SECONDS,
    RecordingSession,
    VideoRecordingResult,
    cleanup_video_segments,
    concatenate_videos,
    get_active_session,
    has_active_session,
    remove_active_session,
    set_active_session,
)

logger = get_logger(__name__)


class AndroidDeviceController(MobileDeviceController):
    def __init__(
        self,
        device_id: str,
        adb_client: AdbClient,
        ui_adb_client: UIAutomatorClient,
        device_width: int,
        device_height: int,
    ):
        self.device_id = device_id
        self.adb_client = adb_client
        self.ui_adb_client = ui_adb_client
        self.device_width = device_width
        self.device_height = device_height
        self._device: AdbDevice | None = None

    @property
    def device(self) -> AdbDevice:
        if self._device is None:
            self._device = self.adb_client.device(serial=self.device_id)
        return self._device

    async def tap(
        self,
        coords: CoordinatesSelectorRequest,
        long_press: bool = False,
        long_press_duration: int = 1000,
    ) -> TapOutput:
        try:
            if long_press:
                cmd = (
                    f"input swipe {coords.x} {coords.y} {coords.x} {coords.y} {long_press_duration}"
                )
            else:
                cmd = f"input tap {coords.x} {coords.y}"

            self.device.shell(cmd)
            return TapOutput(error=None)
        except Exception as e:
            return TapOutput(error=f"ADB tap failed: {str(e)}")

    async def swipe(
        self,
        start: CoordinatesSelectorRequest,
        end: CoordinatesSelectorRequest,
        duration: int = 400,
    ) -> str | None:
        try:
            cmd = f"input touchscreen swipe {start.x} {start.y} {end.x} {end.y} {duration}"
            self.device.shell(cmd)
            return None
        except Exception as e:
            return f"ADB swipe failed: {str(e)}"

    async def get_screen_data(self) -> ScreenDataResponse:
        """Get screen data using the UIAutomator2 client"""
        try:
            logger.info("Using UIAutomator2 for screen data retrieval")
            ui_data = self.ui_adb_client.get_screen_data()
            return ScreenDataResponse(
                base64=ui_data.base64,
                elements=ui_data.elements,
                width=ui_data.width,
                height=ui_data.height,
                platform="android",
            )
        except Exception as e:
            logger.error(f"Failed to get screen data: {e}")
            raise

    async def screenshot(self) -> str:
        try:
            return (await self.get_screen_data()).base64
        except Exception as e:
            logger.error(f"Failed to take screenshot: {e}")
            raise

    async def input_text(self, text: str) -> bool:
        try:
            self.ui_adb_client.send_text(text)
            return True
        except Exception as e:
            logger.warning(f"UIAutomator2 send_text failed: {e}, falling back to ADB shell")
            return self._input_text_adb_fallback(text)

    def _input_text_adb_fallback(self, text: str) -> bool:
        """Fallback method using ADB shell input text command."""
        try:
            parts = text.split("%s")
            for i, part in enumerate(parts):
                to_write = ""
                for char in part:
                    if char == " ":
                        to_write += "%s"
                    elif char in ["&", "<", ">", "|", ";", "(", ")", "$", "`", "\\", '"', "'"]:
                        to_write += f"\\{char}"
                    else:
                        to_write += char

                if to_write:
                    self.device.shell(f"input text '{to_write}'")

                if i < len(parts) - 1:
                    self.device.shell("input keyevent 62")

            return True
        except Exception as e:
            logger.error(f"Failed to input text via ADB fallback: {e}")
            return False

    async def launch_app(self, package_or_bundle_id: str) -> bool:
        try:
            self.device.app_start(package_or_bundle_id)
            return True
        except Exception as e:
            logger.error(f"Failed to launch app {package_or_bundle_id}: {e}")
            return False

    async def terminate_app(self, package_or_bundle_id: str | None) -> bool:
        try:
            if package_or_bundle_id is None:
                current_app = self._get_current_foreground_package()
                if current_app:
                    logger.info(f"Stopping currently running app: {current_app}")
                    self.device.app_stop(current_app)
                else:
                    logger.warning("No foreground app detected")
                    return False
            else:
                self.device.app_stop(package_or_bundle_id)
            return True
        except Exception as e:
            logger.error(f"Failed to terminate app {package_or_bundle_id}: {e}")
            return False

    async def open_url(self, url: str) -> bool:
        try:
            self.device.shell(f"am start -a android.intent.action.VIEW -d {url}")
            return True
        except Exception as e:
            logger.error(f"Failed to open URL {url}: {e}")
            return False

    async def press_back(self) -> bool:
        try:
            self.device.shell("input keyevent 4")
            return True
        except Exception as e:
            logger.error(f"Failed to press back: {e}")
            return False

    async def press_home(self) -> bool:
        try:
            self.device.shell("input keyevent 3")
            return True
        except Exception as e:
            logger.error(f"Failed to press home: {e}")
            return False

    async def press_enter(self) -> bool:
        try:
            self.device.shell("input keyevent 66")
            return True
        except Exception as e:
            logger.error(f"Failed to press enter: {e}")
            return False

    async def get_ui_hierarchy(self) -> list[dict]:
        try:
            device_data = await self.get_screen_data()
            return device_data.elements
        except Exception as e:
            logger.error(f"Failed to get UI hierarchy: {e}")
            return []

    def find_element(
        self,
        ui_hierarchy: list[dict],
        resource_id: str | None = None,
        text: str | None = None,
        index: int = 0,
    ) -> tuple[dict | None, Bounds | None, str | None]:
        if not resource_id and not text:
            return None, None, "No resource_id or text provided"

        matches = []
        for element in ui_hierarchy:
            if resource_id and element.get("resource-id") == resource_id:
                matches.append(element)
            elif text and (element.get("text") == text or element.get("accessibilityText") == text):
                matches.append(element)

        if not matches:
            criteria = f"resource_id='{resource_id}'" if resource_id else f"text='{text}'"
            return None, None, f"No element found with {criteria}"

        if index >= len(matches):
            criteria = f"resource_id='{resource_id}'" if resource_id else f"text='{text}'"
            return (
                None,
                None,
                f"Index {index} out of range for {criteria} (found {len(matches)} matches)",
            )

        element = matches[index]
        bounds = self._extract_bounds(element)

        return element, bounds, None

    def _get_current_foreground_package(self) -> str | None:
        try:
            result = self.device.shell("dumpsys window | grep mCurrentFocus")

            # Convert to string if bytes
            if isinstance(result, bytes):
                result_str = result.decode("utf-8")
            elif isinstance(result, str):
                result_str = result
            else:
                return None

            if result_str and "=" in result_str:
                parts = result_str.split("/")
                if len(parts) > 0:
                    package = parts[0].split()[-1]
                    return package if package else None
            return None
        except Exception as e:
            logger.error(f"Failed to get current foreground package: {e}")
            return None

    def _extract_bounds(self, element: dict) -> Bounds | None:
        bounds_str = element.get("bounds")
        if not bounds_str or not isinstance(bounds_str, str):
            return None

        try:
            match = re.match(r"\[(\d+),(\d+)\]\[(\d+),(\d+)\]", bounds_str)
            if match:
                return Bounds(
                    x1=int(match.group(1)),
                    y1=int(match.group(2)),
                    x2=int(match.group(3)),
                    y2=int(match.group(4)),
                )
        except (ValueError, IndexError):
            return None

        return None

    async def erase_text(self, nb_chars: int | None = None) -> bool:
        try:
            chars_to_delete = nb_chars if nb_chars is not None else 50
            for _ in range(chars_to_delete):
                self.device.shell("input keyevent KEYCODE_DEL")
            return True
        except Exception as e:
            logger.error(f"Failed to erase text: {e}")
            return False

    async def cleanup(self) -> None:
        pass

    def get_compressed_b64_screenshot(self, image_base64: str, quality: int = 50) -> str:
        if image_base64.startswith("data:image"):
            image_base64 = image_base64.split(",")[1]

        image_data = base64.b64decode(image_base64)
        image = Image.open(BytesIO(image_data))

        compressed_io = BytesIO()
        image.save(compressed_io, format="JPEG", quality=quality, optimize=True)

        compressed_base64 = base64.b64encode(compressed_io.getvalue()).decode("utf-8")
        return compressed_base64

    async def _start_android_segment(
        self, session: RecordingSession
    ) -> asyncio.subprocess.Process | None:
        """Start a single Android recording segment."""
        segment_path = f"/sdcard/screen_recording_{session.android_segment_index}.mp4"
        cmd = f"screenrecord --time-limit {ANDROID_MAX_RECORDING_DURATION_SECONDS} {segment_path}"

        process = await asyncio.create_subprocess_shell(
            f'adb -s {self.device_id} shell "{cmd}"',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        session.android_device_path = segment_path
        session.process = process
        return process

    async def _android_auto_restart_loop(
        self,
        session: RecordingSession,
        max_duration_seconds: int,
    ) -> None:
        """Background task that auto-restarts Android recording when segment ends."""
        total_elapsed = 0

        while total_elapsed < max_duration_seconds:
            process = session.process
            if process is None:
                break

            try:
                await asyncio.wait_for(
                    process.wait(),
                    timeout=ANDROID_MAX_RECORDING_DURATION_SECONDS + 5,
                )
            except TimeoutError:
                pass
            except asyncio.CancelledError:
                return

            if not has_active_session(self.device_id):
                return

            total_elapsed = time.time() - session.start_time
            if total_elapsed >= max_duration_seconds:
                break

            await asyncio.sleep(VIDEO_READY_DELAY_SECONDS)

            temp_dir = tempfile.mkdtemp(prefix="mobile_use_video_segment_")
            local_segment = Path(temp_dir) / f"segment_{session.android_segment_index}.mp4"

            try:
                self.device.sync.pull(session.android_device_path, str(local_segment))
                self.device.shell(f"rm -f {session.android_device_path}")
                session.android_video_segments.append(local_segment)
                logger.info(
                    f"Saved Android segment {session.android_segment_index} to {local_segment}"
                )
            except Exception as e:
                error_msg = f"Failed to pull segment {session.android_segment_index}: {e}"
                logger.warning(error_msg)
                session.errors.append(error_msg)

            session.android_segment_index += 1

            try:
                await self._start_android_segment(session)
                logger.info(
                    f"Auto-restarted Android recording (segment {session.android_segment_index})"
                )
            except Exception as e:
                error_msg = f"Failed to restart Android recording: {e}"
                logger.error(error_msg)
                session.errors.append(error_msg)
                break

    async def start_video_recording(
        self,
        max_duration_seconds: int = DEFAULT_MAX_DURATION_SECONDS,
    ) -> VideoRecordingResult:
        """Start screen recording on Android device using adb shell screenrecord."""
        if has_active_session(self.device_id):
            return VideoRecordingResult(
                success=False,
                message=f"Recording already in progress for device {self.device_id}",
            )

        try:
            session = RecordingSession(
                device_id=self.device_id,
                start_time=time.time(),
                android_video_segments=[],
                android_segment_index=0,
            )

            set_active_session(self.device_id, session)
            await self._start_android_segment(session)

            restart_task = asyncio.create_task(
                self._android_auto_restart_loop(session, max_duration_seconds)
            )
            session.android_restart_task = restart_task

            logger.info(f"Started Android screen recording on {self.device_id}")
            return VideoRecordingResult(
                success=True,
                message=(
                    f"Recording started (max {max_duration_seconds}s, "
                    f"auto-restarts every {ANDROID_MAX_RECORDING_DURATION_SECONDS}s)."
                ),
            )

        except Exception as e:
            logger.error(f"Failed to start Android recording: {e}")
            remove_active_session(self.device_id)
            return VideoRecordingResult(
                success=False,
                message=f"Failed to start recording: {e}",
            )

    async def stop_video_recording(self) -> VideoRecordingResult:
        """Stop Android recording, pull all segments, and concatenate them."""
        session = get_active_session(self.device_id)

        if not session:
            return VideoRecordingResult(
                success=False,
                message=f"No active recording for device {self.device_id}",
            )

        try:
            if session.android_restart_task:
                session.android_restart_task.cancel()
                try:
                    await session.android_restart_task
                except asyncio.CancelledError:
                    pass

            process = session.process
            if process is not None:
                try:
                    process.terminate()
                    await asyncio.wait_for(process.wait(), timeout=5.0)
                except TimeoutError:
                    process.kill()
                    await process.wait()

            self.device.shell("pkill -2 screenrecord || true")
            await asyncio.sleep(VIDEO_READY_DELAY_SECONDS)

            temp_dir = tempfile.mkdtemp(prefix="mobile_use_video_")
            final_segment = Path(temp_dir) / f"segment_{session.android_segment_index}.mp4"

            try:
                self.device.sync.pull(session.android_device_path, str(final_segment))
                self.device.shell(f"rm -f {session.android_device_path}")
                session.android_video_segments.append(final_segment)
            except Exception as e:
                logger.warning(f"Failed to pull final segment: {e}")

            for i in range(session.android_segment_index + 1):
                self.device.shell(f"rm -f /sdcard/screen_recording_{i}.mp4")

            all_segments = session.android_video_segments

            if not all_segments:
                remove_active_session(self.device_id)
                return VideoRecordingResult(
                    success=False,
                    message="No video segments were captured",
                )

            output_dir = tempfile.mkdtemp(prefix="mobile_use_video_final_")
            output_path = Path(output_dir) / "recording.mp4"

            if len(all_segments) == 1:
                all_segments[0].rename(output_path)
            else:
                success = await concatenate_videos(all_segments, output_path)
                if not success:
                    output_path = all_segments[-1]
                    logger.warning("Concatenation failed, using last segment only")

            cleanup_video_segments(all_segments, keep_path=output_path)

            errors = session.errors.copy()
            remove_active_session(self.device_id)

            duration = time.time() - session.start_time
            segment_count = len(all_segments)
            logger.info(
                f"Stopped Android recording after {duration:.1f}s, "
                f"{segment_count} segment(s), saved to {output_path}"
            )

            message = f"Recording stopped after {duration:.1f}s ({segment_count} segments)"
            if errors:
                message += f". Warnings during recording: {'; '.join(errors)}"

            return VideoRecordingResult(
                success=True,
                message=message,
                video_path=output_path,
            )

        except Exception as e:
            logger.error(f"Failed to stop Android recording: {e}")
            remove_active_session(self.device_id)
            return VideoRecordingResult(
                success=False,
                message=f"Failed to stop recording: {e}",
            )



from pydantic import BaseModel, ConfigDict, Field


class TapOutput(BaseModel):
    """Output from tap operations."""

    error: str | None = Field(default=None, description="Error message if tap failed")


class Bounds(BaseModel):
    """Represents the bounds of a UI element."""

    x1: int
    y1: int
    x2: int
    y2: int

    def get_center(self) -> "CoordinatesSelectorRequest":
        """Get the center point of the bounds."""
        return CoordinatesSelectorRequest(
            x=(self.x1 + self.x2) // 2,
            y=(self.y1 + self.y2) // 2,
        )


class CoordinatesSelectorRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")
    x: int
    y: int

    def to_str(self):
        return f"{self.x}, {self.y}"


class PercentagesSelectorRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")
    """
    0%,0%        # top-left corner
    100%,100%    # bottom-right corner
    50%,50%      # center
    """

    x_percent: int = Field(ge=0, le=100, description="X percentage (0-100)")
    y_percent: int = Field(ge=0, le=100, description="Y percentage (0-100)")

    def to_str(self):
        return f"{self.x_percent}%, {self.y_percent}%"

    def to_coords(self, width: int, height: int) -> CoordinatesSelectorRequest:
        """Convert percentages to pixel coordinates."""
        x = min(max(int(width * self.x_percent / 100), 0), max(0, width - 1))
        y = min(max(int(height * self.y_percent / 100), 0), max(0, height - 1))
        return CoordinatesSelectorRequest(x=x, y=y)


class SwipeStartEndCoordinatesRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")
    start: CoordinatesSelectorRequest
    end: CoordinatesSelectorRequest

    def to_dict(self):
        return {"start": self.start.to_str(), "end": self.end.to_str()}


class SwipeStartEndPercentagesRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")
    start: PercentagesSelectorRequest
    end: PercentagesSelectorRequest

    def to_dict(self):
        return {"start": self.start.to_str(), "end": self.end.to_str()}

    def to_coords(self, width: int, height: int) -> SwipeStartEndCoordinatesRequest:
        """Convert percentage-based swipe to coordinate-based swipe."""
        return SwipeStartEndCoordinatesRequest(
            start=self.start.to_coords(width, height),
            end=self.end.to_coords(width, height),
        )


class SwipeRequest(BaseModel):
    """
    Swipe from start to end position using coordinates or percentages.
    """

    model_config = ConfigDict(extra="forbid")
    swipe_mode: SwipeStartEndCoordinatesRequest | SwipeStartEndPercentagesRequest = Field(
        description="Start and end positions. Use EITHER (x, y) OR (x_percent, y_percent)."
    )
    duration: int | None = Field(
        default=None,
        description="Swipe duration in ms. If not provided, tool functions default to 400ms.",
        ge=1,
        le=10000,
    )

    def to_dict(self):
        res = {}
        if isinstance(
            self.swipe_mode,
            SwipeStartEndCoordinatesRequest | SwipeStartEndPercentagesRequest,
        ):
            res |= self.swipe_mode.to_dict()
        if self.duration:
            res |= {"duration": self.duration}
        return res



from typing import Annotated

from langchain_core.messages import ToolMessage
from langchain_core.tools import BaseTool, tool
from langchain_core.tools.base import InjectedToolCallId
from langgraph.prebuilt import InjectedState
from langgraph.types import Command
from pydantic import BaseModel

from minitap.mobile_use.constants import EXECUTOR_MESSAGES_KEY
from minitap.mobile_use.context import MobileUseContext
from minitap.mobile_use.controllers.controller_factory import create_device_controller
from minitap.mobile_use.controllers.unified_controller import UnifiedMobileController
from minitap.mobile_use.graph.state import State
from minitap.mobile_use.tools.tool_wrapper import ToolWrapper
from minitap.mobile_use.tools.types import Target
from minitap.mobile_use.tools.utils import focus_element_if_needed, move_cursor_to_end_if_bounds
from minitap.mobile_use.utils.logger import get_logger
from minitap.mobile_use.utils.ui_hierarchy import (
    find_element_by_resource_id,
    get_element_text,
    text_input_is_empty,
)

logger = get_logger(__name__)

MAX_CLEAR_TRIES = 5
DEFAULT_CHARS_TO_ERASE = 50


class ClearTextResult(BaseModel):
    success: bool
    error_message: str | None
    chars_erased: int
    final_text: str | None


class TextClearer:
    def __init__(self, ctx: MobileUseContext, state: State):
        self.ctx = ctx
        self.state = state

    async def _refresh_ui_hierarchy(self) -> None:
        device_controller = create_device_controller(self.ctx)
        screen_data = await device_controller.get_screen_data()
        self.state.latest_ui_hierarchy = screen_data.elements

    async def _get_element_info(
        self, resource_id: str | None
    ) -> tuple[object | None, str | None, str | None]:
        if not self.state.latest_ui_hierarchy:
            await self._refresh_ui_hierarchy()

        if not self.state.latest_ui_hierarchy:
            return None, None, None

        element = None
        if resource_id:
            element = find_element_by_resource_id(
                ui_hierarchy=self.state.latest_ui_hierarchy, resource_id=resource_id
            )

        if not element:
            return None, None, None

        current_text = get_element_text(element)
        hint_text = get_element_text(element, hint_text=True)

        return element, current_text, hint_text

    def _format_text_with_hint_info(self, text: str | None, hint_text: str | None) -> str | None:
        if text is None:
            return None

        is_hint_text = hint_text is not None and hint_text != "" and hint_text == text

        if is_hint_text:
            return f"{text} (which is the hint text, the input is very likely empty)"

        return text

    def _should_clear_text(self, current_text: str | None, hint_text: str | None) -> bool:
        return current_text is not None and current_text != "" and current_text != hint_text

    async def _prepare_element_for_clearing(
        self,
        target: Target,
    ) -> bool:
        if not await focus_element_if_needed(
            ctx=self.ctx,
            target=target,
        ):
            return False

        await move_cursor_to_end_if_bounds(
            ctx=self.ctx,
            state=self.state,
            target=target,
        )
        return True

    async def _erase_text_attempt(self, text_length: int) -> str | None:
        chars_to_erase = text_length + 1
        logger.info(f"Erasing {chars_to_erase} characters from the input")

        controller = UnifiedMobileController(self.ctx)
        result = await controller.erase_text()

        if not result:
            logger.error("Failed to erase text")
            return "Failed to erase text"

        return None

    async def _clear_with_retries(
        self,
        target: Target,
        initial_text: str,
        hint_text: str | None,
    ) -> tuple[bool, str | None, int]:
        current_text = initial_text
        erased_chars = 0

        for attempt in range(1, MAX_CLEAR_TRIES + 1):
            logger.info(f"Clear attempt {attempt}/{MAX_CLEAR_TRIES}")

            chars_to_erase = len(current_text) if current_text else DEFAULT_CHARS_TO_ERASE
            error = await self._erase_text_attempt(text_length=chars_to_erase)

            if error:
                return False, current_text, 0
            erased_chars += chars_to_erase

            await self._refresh_ui_hierarchy()
            elt = None
            if target.resource_id:
                elt = find_element_by_resource_id(
                    ui_hierarchy=self.state.latest_ui_hierarchy or [],
                    resource_id=target.resource_id,
                )
                if elt:
                    current_text = get_element_text(elt)
                    logger.info(f"Current text: {current_text}")
                    if text_input_is_empty(text=current_text, hint_text=hint_text):
                        break

            await move_cursor_to_end_if_bounds(
                ctx=self.ctx,
                state=self.state,
                target=target,
                elt=elt,
            )

        return True, current_text, erased_chars

    def _create_result(
        self,
        success: bool,
        error_message: str | None,
        chars_erased: int,
        final_text: str | None,
        hint_text: str | None,
    ) -> ClearTextResult:
        formatted_final_text = self._format_text_with_hint_info(final_text, hint_text)

        return ClearTextResult(
            success=success,
            error_message=error_message,
            chars_erased=chars_erased,
            final_text=formatted_final_text,
        )

    def _handle_no_clearing_needed(
        self, current_text: str | None, hint_text: str | None
    ) -> ClearTextResult:
        return self._create_result(
            success=True,
            error_message=None,
            chars_erased=-1,
            final_text=current_text,
            hint_text=hint_text,
        )

    async def _handle_element_not_found(
        self, target: Target, hint_text: str | None
    ) -> ClearTextResult:
        if not await self._prepare_element_for_clearing(target=target):
            return self._create_result(
                success=False,
                error_message="Failed to focus element",
                chars_erased=0,
                final_text=None,
                hint_text=None,
            )

        controller = UnifiedMobileController(self.ctx)
        output = await controller.erase_text()
        await self._refresh_ui_hierarchy()

        _, final_text, _ = await self._get_element_info(target.resource_id)

        return self._create_result(
            success=output,
            error_message="Erase text failed" if not output else None,
            chars_erased=0,  # Unknown since we don't have initial text
            final_text=final_text,
            hint_text=hint_text,
        )

    async def clear_input_text(
        self,
        target: Target,
    ) -> ClearTextResult:
        element, current_text, hint_text = await self._get_element_info(
            resource_id=target.resource_id,
        )

        if not element:
            return await self._handle_element_not_found(target=target, hint_text=hint_text)

        if not self._should_clear_text(current_text, hint_text):
            return self._handle_no_clearing_needed(current_text, hint_text)

        if not await self._prepare_element_for_clearing(target=target):
            return self._create_result(
                success=False,
                error_message="Failed to focus element",
                chars_erased=0,
                final_text=current_text,
                hint_text=hint_text,
            )

        success, final_text, chars_erased = await self._clear_with_retries(
            target=target,
            initial_text=current_text or "",
            hint_text=hint_text,
        )

        error_message = None if success else "Failed to clear text after retries"

        return self._create_result(
            success=success,
            error_message=error_message,
            chars_erased=chars_erased,
            final_text=final_text,
            hint_text=hint_text,
        )


def get_focus_and_clear_text_tool(ctx: MobileUseContext) -> BaseTool:
    @tool
    async def focus_and_clear_text(
        agent_thought: str,
        target: Target,
        tool_call_id: Annotated[str, InjectedToolCallId],
        state: Annotated[State, InjectedState],
    ):
        """
        Clears all the text from the text field, by focusing it if needed.

        Args:
            agent_thought: The thought of the agent.
            target: The target text field to clear.
        """
        clearer = TextClearer(ctx, state)
        result = await clearer.clear_input_text(target=target)

        agent_outcome = (
            focus_and_clear_text_wrapper.on_failure_fn(result.error_message)
            if not result.success
            else focus_and_clear_text_wrapper.on_success_fn(
                nb_char_erased=result.chars_erased, new_text_value=result.final_text
            )
        )

        tool_message = ToolMessage(
            tool_call_id=tool_call_id,
            content=agent_outcome,
            additional_kwargs={"error": result.error_message} if not result.success else {},
            status="error" if not result.success else "success",
        )

        return Command(
            update=await state.asanitize_update(
                ctx=ctx,
                update={
                    "agents_thoughts": [agent_thought, agent_outcome],
                    EXECUTOR_MESSAGES_KEY: [tool_message],
                },
                agent="executor",
            ),
        )

    return focus_and_clear_text


def _format_success_message(nb_char_erased: int, new_text_value: str | None) -> str:
    if nb_char_erased == -1:
        msg = "No text clearing was needed (the input was already empty)."
    else:
        msg = f"Text erased successfully. {nb_char_erased} characters were erased."

    if new_text_value is not None:
        msg += f" New text in the input is '{new_text_value}'."

    return msg


def _format_failure_message(output: str | None) -> str:
    return "Failed to erase text. " + (str(output) if output else "")


focus_and_clear_text_wrapper = ToolWrapper(
    tool_fn_getter=get_focus_and_clear_text_tool,
    on_success_fn=_format_success_message,
    on_failure_fn=_format_failure_message,
)



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, SecretStr


class BrowserStackClientConfig(BaseModel):
    model_config = ConfigDict(frozen=True)
    username: str
    access_key: SecretStr
    device_name: str
    platform_version: str
    app_url: str
    hub_url: str | None = None
    project_name: str | None = None
    build_name: str | None = None
    session_name: str | None = None

    @classmethod
    def with_overrides(
        cls,
        username: str | None = None,
        access_key: str | None = None,
        device_name: str | None = None,
        platform_version: str | None = None,
        app_url: str | None = None,
        hub_url: str | None = None,
        project_name: str | None = None,
        build_name: str | None = None,
        session_name: str | None = None,
        base: BrowserStackClientConfig | None = None,
    ) -> BrowserStackClientConfig:
        """Create a BrowserStackClientConfig with only specified fields overridden."""
        if base is None:
            raise ValueError("base config is required for BrowserStackClientConfig.with_overrides")
        overrides = {
            k: v
            for k, v in {
                "username": username,
                "access_key": access_key,
                "device_name": device_name,
                "platform_version": platform_version,
                "app_url": app_url,
                "hub_url": hub_url,
                "project_name": project_name,
                "build_name": build_name,
                "session_name": session_name,
            }.items()
            if v is not None
        }
        if not overrides:
            return base
        return base.model_copy(update=overrides)


class WdaClientConfig(BaseModel):
    model_config = ConfigDict(frozen=True)
    wda_url: str = "http://localhost:8100"
    timeout: float = 30.0
    auto_start_iproxy: bool = True
    auto_start_wda: bool = True
    wda_project_path: str | None = None
    wda_startup_timeout: float = 120.0

    @classmethod
    def with_overrides(
        cls,
        wda_url: str | None = None,
        timeout: float | None = None,
        auto_start_iproxy: bool | None = None,
        auto_start_wda: bool | None = None,
        wda_project_path: str | None = None,
        wda_startup_timeout: float | None = None,
    ) -> WdaClientConfig:
        """Create a WdaClientConfig with only specified fields overridden.

        Example:
            config = WdaClientConfig.with_overrides(
                wda_url="http://localhost:8101",
                auto_start_wda=False,
            )
        """
        base = cls()
        overrides = {
            k: v
            for k, v in {
                "wda_url": wda_url,
                "timeout": timeout,
                "auto_start_iproxy": auto_start_iproxy,
                "auto_start_wda": auto_start_wda,
                "wda_project_path": wda_project_path,
                "wda_startup_timeout": wda_startup_timeout,
            }.items()
            if v is not None
        }
        if not overrides:
            return base
        return base.model_copy(update=overrides)


class IdbClientConfig(BaseModel):
    model_config = ConfigDict(frozen=True)
    host: str | None = None
    port: int | None = None

    @classmethod
    def with_overrides(
        cls,
        host: str | None = None,
        port: int | None = None,
    ) -> IdbClientConfig:
        """Create an IdbClientConfig with only specified fields overridden."""
        base = cls()
        overrides = {k: v for k, v in {"host": host, "port": port}.items() if v is not None}
        if not overrides:
            return base
        return base.model_copy(update=overrides)


class IosClientConfig(BaseModel):
    model_config = ConfigDict(frozen=True)
    wda: WdaClientConfig = WdaClientConfig()
    idb: IdbClientConfig = IdbClientConfig()
    browserstack: BrowserStackClientConfig | None = None

    @classmethod
    def with_overrides(
        cls,
        wda: WdaClientConfig | None = None,
        idb: IdbClientConfig | None = None,
        browserstack: BrowserStackClientConfig | None = None,
    ) -> IosClientConfig:
        """Create an IosClientConfig with only specified fields overridden."""
        base = cls()
        overrides = {
            k: v
            for k, v in {"wda": wda, "idb": idb, "browserstack": browserstack}.items()
            if v is not None
        }
        if not overrides:
            return base
        return base.model_copy(update=overrides)



from enum import Enum
from typing import Annotated

from langchain_core.messages import ToolMessage
from langchain_core.tools import tool
from langchain_core.tools.base import InjectedToolCallId
from langgraph.prebuilt import InjectedState
from langgraph.types import Command
from pydantic import BeforeValidator

from minitap.mobile_use.constants import EXECUTOR_MESSAGES_KEY
from minitap.mobile_use.context import MobileUseContext
from minitap.mobile_use.controllers.unified_controller import UnifiedMobileController
from minitap.mobile_use.graph.state import State
from minitap.mobile_use.tools.tool_wrapper import ToolWrapper


class Key(Enum):
    ENTER = "Enter"
    HOME = "Home"
    BACK = "Back"


def normalize_key(value: str | Key) -> str:
    """Convert key input to Title Case for case-insensitive matching."""
    if isinstance(value, Key):
        return value.value
    return value.title()


CaseInsensitiveKey = Annotated[Key, BeforeValidator(normalize_key)]


def get_press_key_tool(ctx: MobileUseContext):
    @tool
    async def press_key(
        agent_thought: str,
        key: CaseInsensitiveKey,
        tool_call_id: Annotated[str, InjectedToolCallId],
        state: Annotated[State, InjectedState],
    ) -> Command:
        """Press a key on the device."""
        controller = UnifiedMobileController(ctx)
        match key:
            case Key.HOME:
                output = await controller.go_home()
            case Key.BACK:
                output = await controller.go_back()
            case Key.ENTER:
                output = await controller.press_enter()
        has_failed = not output

        agent_outcome = (
            press_key_wrapper.on_failure_fn(key)
            if has_failed
            else press_key_wrapper.on_success_fn(key)
        )

        tool_message = ToolMessage(
            tool_call_id=tool_call_id,
            content=agent_outcome,
            additional_kwargs={"error": output} if has_failed else {},
            status="error" if has_failed else "success",
        )
        return Command(
            update=await state.asanitize_update(
                ctx=ctx,
                update={
                    "agents_thoughts": [agent_thought, agent_outcome],
                    EXECUTOR_MESSAGES_KEY: [tool_message],
                },
                agent="executor",
            ),
        )

    return press_key


press_key_wrapper = ToolWrapper(
    tool_fn_getter=get_press_key_tool,
    on_success_fn=lambda key: f"Key {key.value} pressed successfully.",
    on_failure_fn=lambda key: f"Failed to press key {key.value}.",
)



import logging
import sys
from enum import Enum
from pathlib import Path

from colorama import Fore, Style, init

init(autoreset=True)


class LogLevel(Enum):
    DEBUG = ("DEBUG", Fore.MAGENTA, "🔍")
    INFO = ("INFO", Fore.WHITE, "ℹ")
    SUCCESS = ("SUCCESS", Fore.GREEN, "✓")
    WARNING = ("WARNING", Fore.YELLOW, "⚠")
    ERROR = ("ERROR", Fore.RED, "❌")
    CRITICAL = ("CRITICAL", Fore.RED + Style.BRIGHT, "💥")


class MobileUseLogger:
    def __init__(
        self,
        name: str,
        log_file: str | Path | None = None,
        console_level: str = "INFO",
        file_level: str = "DEBUG",
        enable_file_logging: bool = True,
    ):
        """
        Initialize the MobileUse logger.

        Args:
            name: Logger name (usually __name__)
            log_file: Path to log file (defaults to logs/{name}.log)
            console_level: Minimum level for console output
            file_level: Minimum level for file output
            enable_file_logging: Whether to enable file logging
        """
        self.name = name
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)

        self.logger.handlers.clear()

        self._setup_console_handler(console_level)

        if enable_file_logging:
            self._setup_file_handler(log_file, file_level)

    def _setup_console_handler(self, level: str):
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setLevel(getattr(logging, level.upper()))

        console_formatter = ColoredFormatter()
        console_handler.setFormatter(console_formatter)

        self.logger.addHandler(console_handler)

    def _setup_file_handler(self, log_file: str | Path | None, level: str):
        if log_file is None:
            log_file = Path("logs") / f"{self.name.replace('.', '_')}.log"

        log_file = Path(log_file)
        log_file.parent.mkdir(parents=True, exist_ok=True)

        file_handler = logging.FileHandler(log_file, encoding="utf-8")
        file_handler.setLevel(getattr(logging, level.upper()))

        file_formatter = logging.Formatter(
            fmt="%(asctime)s | %(name)s | %(levelname)s | %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        file_handler.setFormatter(file_formatter)

        self.logger.addHandler(file_handler)

    def debug(self, message: str, **kwargs):
        self.logger.debug(message, extra={"log_level": LogLevel.DEBUG}, **kwargs)

    def info(self, message: str, **kwargs):
        self.logger.info(message, extra={"log_level": LogLevel.INFO}, **kwargs)

    def success(self, message: str, **kwargs):
        self.logger.info(message, extra={"log_level": LogLevel.SUCCESS}, **kwargs)

    def warning(self, message: str, **kwargs):
        self.logger.warning(message, extra={"log_level": LogLevel.WARNING}, **kwargs)

    def error(self, message: str, **kwargs):
        self.logger.error(message, extra={"log_level": LogLevel.ERROR}, **kwargs)

    def critical(self, message: str, **kwargs):
        self.logger.critical(message, extra={"log_level": LogLevel.CRITICAL}, **kwargs)

    def header(self, message: str, **_kwargs):
        separator = "=" * 60
        colored_separator = f"{Fore.CYAN}{separator}{Style.RESET_ALL}"
        colored_message = f"{Fore.CYAN}{message}{Style.RESET_ALL}"

        print(colored_separator)
        print(colored_message)
        print(colored_separator)
        self.logger.info(f"\n{separator}\n{message}\n{separator}")


class ColoredFormatter(logging.Formatter):
    def format(self, record):
        log_level = getattr(record, "log_level", LogLevel.INFO)
        _level_name, color, symbol = log_level.value

        colored_message = f"{color}{symbol} {record.getMessage()}{Style.RESET_ALL}"

        return colored_message


_loggers = {}


def get_logger(
    name: str,
    log_file: str | Path | None = None,
    console_level: str = "INFO",
    file_level: str = "DEBUG",
    enable_file_logging: bool = False,
) -> MobileUseLogger:
    """
    Get or create a logger instance.

    Args:
        name: Logger name (usually __name__)
        log_file: Path to log file (defaults to logs/{name}.log)
        console_level: Minimum level for console output
        file_level: Minimum level for file output
        enable_file_logging: Whether to enable file logging

    Returns:
        MobileUseLogger instance
    """
    if name not in _loggers:
        _loggers[name] = MobileUseLogger(
            name=name,
            log_file=log_file,
            console_level=console_level,
            file_level=file_level,
            enable_file_logging=enable_file_logging,
        )

    return _loggers[name]


def log_debug(message: str, logger_name: str = "mobile-use"):
    get_logger(logger_name).debug(message)


def log_info(message: str, logger_name: str = "mobile-use"):
    get_logger(logger_name).info(message)


def log_success(message: str, logger_name: str = "mobile-use"):
    get_logger(logger_name).success(message)


def log_warning(message: str, logger_name: str = "mobile-use"):
    get_logger(logger_name).warning(message)


def log_error(message: str, logger_name: str = "mobile-use"):
    get_logger(logger_name).error(message)


def log_critical(message: str, logger_name: str = "mobile-use"):
    get_logger(logger_name).critical(message)


def log_header(message: str, logger_name: str = "mobile-use"):
    get_logger(logger_name).header(message)


def get_server_logger() -> MobileUseLogger:
    return get_logger(
        name="mobile-use.servers",
        console_level="INFO",
        file_level="DEBUG",
    )



import requests
from minitap.mobile_use.utils.logger import get_logger

logger = get_logger(__name__)


def curl_from_request(req: requests.PreparedRequest) -> str:
    """Converts a requests.PreparedRequest object to a valid cURL command string."""
    command = ["curl", f"-X {req.method}"]

    for key, value in req.headers.items():
        command.append(f'-H "{key}: {value}"')

    if req.body:
        body = req.body
        if isinstance(body, bytes):
            body = body.decode("utf-8")
        # Escape single quotes in the body for shell safety
        body = body.replace("'", "'\\''")
        command.append(f"-d '{body}'")

    command.append(f"'{req.url}'")

    return " ".join(command)


def logging_hook(response, *args, **kwargs):
    """Hook to log the request as a cURL command."""
    curl_command = curl_from_request(response.request)
    logger.debug(f"\n--- cURL Command ---\n{curl_command}\n--------------------")


def get_session_with_curl_logging() -> requests.Session:
    """Returns a requests.Session with cURL logging enabled."""
    session = requests.Session()
    session.hooks["response"] = [logging_hook]
    return session



import random
import string

from minitap.mobile_use.agents.planner.types import Subgoal, SubgoalStatus
from datetime import datetime, UTC


def get_current_subgoal(subgoals: list[Subgoal]) -> Subgoal | None:
    return next((s for s in subgoals if s.status == SubgoalStatus.PENDING), None)


def get_subgoals_by_ids(subgoals: list[Subgoal], ids: list[str]) -> list[Subgoal]:
    return [s for s in subgoals if s.id in ids]


def get_next_subgoal(subgoals: list[Subgoal]) -> Subgoal | None:
    return next((s for s in subgoals if s.status == SubgoalStatus.NOT_STARTED), None)


def nothing_started(subgoals: list[Subgoal]) -> bool:
    return all(s.status == SubgoalStatus.NOT_STARTED for s in subgoals)


def complete_current_subgoal(subgoals: list[Subgoal]) -> list[Subgoal]:
    current_subgoal = get_current_subgoal(subgoals)
    if not current_subgoal:
        return subgoals
    current_subgoal.status = SubgoalStatus.SUCCESS
    current_subgoal.ended_at = datetime.now(UTC)
    return subgoals


def complete_subgoals_by_ids(subgoals: list[Subgoal], ids: list[str]) -> list[Subgoal]:
    for subgoal in subgoals:
        if subgoal.id in ids:
            subgoal.status = SubgoalStatus.SUCCESS
            subgoal.ended_at = datetime.now(UTC)
    return subgoals


def fail_current_subgoal(subgoals: list[Subgoal]) -> list[Subgoal]:
    current_subgoal = get_current_subgoal(subgoals)
    if not current_subgoal:
        return subgoals
    current_subgoal.status = SubgoalStatus.FAILURE
    current_subgoal.ended_at = datetime.now(UTC)
    return subgoals


def all_completed(subgoals: list[Subgoal]) -> bool:
    return all(s.status == SubgoalStatus.SUCCESS for s in subgoals)


def one_of_them_is_failure(subgoals: list[Subgoal]) -> bool:
    return any(s.status == SubgoalStatus.FAILURE for s in subgoals)


def start_next_subgoal(subgoals: list[Subgoal]) -> list[Subgoal]:
    next_subgoal = get_next_subgoal(subgoals)
    if not next_subgoal:
        return subgoals
    next_subgoal.status = SubgoalStatus.PENDING
    next_subgoal.started_at = datetime.now(UTC)
    return subgoals


def generate_id(length: int = 6) -> str:
    """Generates a small and distinct random string ID."""
    chars = string.ascii_lowercase + string.digits
    return "".join(random.choice(chars) for _ in range(length))



"""iOS-specific device controller implementation using IDB or WDA."""

import asyncio
import base64
import re
import tempfile
import time
from io import BytesIO
from pathlib import Path

from idb.common.types import HIDButtonType
from PIL import Image

from minitap.mobile_use.clients.idb_client import IdbClientWrapper, IOSAppInfo
from minitap.mobile_use.clients.ios_client import IosClientWrapper
from minitap.mobile_use.controllers.device_controller import (
    MobileDeviceController,
    ScreenDataResponse,
)
from minitap.mobile_use.controllers.types import Bounds, CoordinatesSelectorRequest, TapOutput
from minitap.mobile_use.utils.logger import get_logger
from minitap.mobile_use.utils.video import (
    DEFAULT_MAX_DURATION_SECONDS,
    VIDEO_READY_DELAY_SECONDS,
    RecordingSession,
    VideoRecordingResult,
    get_active_session,
    has_active_session,
    remove_active_session,
    set_active_session,
)

logger = get_logger(__name__)


class iOSDeviceController(MobileDeviceController):
    """iOS device controller using IDB (simulators) or WDA (physical devices)."""

    def __init__(
        self,
        ios_client: IosClientWrapper,
        device_id: str,
        device_width: int,
        device_height: int,
    ):
        self.ios_client = ios_client
        self.device_id = device_id
        self.device_width = device_width
        self.device_height = device_height
        self._is_idb = isinstance(ios_client, IdbClientWrapper)

    async def tap(
        self,
        coords: CoordinatesSelectorRequest,
        long_press: bool = False,
        long_press_duration: int = 1000,
    ) -> TapOutput:
        """Tap at specific coordinates using IDB."""
        try:
            duration = long_press_duration / 1000.0 if long_press else None
            await self.ios_client.tap(x=coords.x, y=coords.y, duration=duration)  # type: ignore[call-arg]
            return TapOutput(error=None)
        except Exception as e:
            return TapOutput(error=f"IDB tap failed: {str(e)}")

    async def swipe(
        self,
        start: CoordinatesSelectorRequest,
        end: CoordinatesSelectorRequest,
        duration: int = 400,
    ) -> str | None:
        """Swipe from start to end coordinates using IDB."""
        try:
            # IDB delta is the number of steps, approximating from duration
            ms_duration_to_percentage = duration / 1000.0
            await self.ios_client.swipe(  # type: ignore[call-arg]
                x_start=start.x,
                y_start=start.y,
                x_end=end.x,
                y_end=end.y,
                duration=ms_duration_to_percentage,
            )
            return None
        except Exception as e:
            return f"IDB swipe failed: {str(e)}"

    async def get_screen_data(self) -> ScreenDataResponse:
        """Get screen data (screenshot and hierarchy in parallel)."""
        try:
            # Run screenshot and hierarchy fetch in parallel
            screenshot_bytes, accessibility_info = await asyncio.gather(
                self.ios_client.screenshot(),  # type: ignore[call-arg]
                self.ios_client.describe_all(),
            )

            if screenshot_bytes is None:
                raise RuntimeError("Screenshot returned None")

            elements = (
                self._process_flat_ios_hierarchy(accessibility_info) if accessibility_info else []
            )

            base64_screenshot = base64.b64encode(screenshot_bytes).decode("utf-8")

            return ScreenDataResponse(
                base64=base64_screenshot,
                elements=elements,
                width=self.device_width,
                height=self.device_height,
                platform="ios",
            )
        except Exception as e:
            logger.error(f"Failed to get screen data: {e}")
            raise

    async def screenshot(self) -> str:
        """Take a screenshot using IDB and return base64 encoded string."""
        try:
            screenshot_bytes = await self.ios_client.screenshot()  # type: ignore[call-arg]
            if screenshot_bytes is None:
                raise RuntimeError("Screenshot returned None")
            return base64.b64encode(screenshot_bytes).decode("utf-8")
        except Exception as e:
            logger.error(f"Failed to take screenshot: {e}")
            raise

    async def input_text(self, text: str) -> bool:
        """Input text using IDB."""
        try:
            return await self.ios_client.text(text)  # type: ignore[call-arg]
        except Exception as e:
            logger.error(f"Failed to input text: {e}")
            return False

    async def launch_app(self, package_or_bundle_id: str) -> bool:
        """Launch an iOS app using IDB."""
        try:
            return await self.ios_client.launch(bundle_id=package_or_bundle_id)  # type: ignore[call-arg]
        except Exception as e:
            logger.error(f"Failed to launch app {package_or_bundle_id}: {e}")
            return False

    async def terminate_app(self, package_or_bundle_id: str | None) -> bool:
        """Terminate an iOS app using IDB."""
        if package_or_bundle_id is None:
            logger.warning("Cannot terminate app: bundle_id is None")
            return False
        try:
            return await self.ios_client.terminate(bundle_id=package_or_bundle_id)  # type: ignore[call-arg]
        except Exception as e:
            logger.error(f"Failed to terminate app {package_or_bundle_id}: {e}")
            return False

    async def open_url(self, url: str) -> bool:
        """Open a URL using IDB."""
        try:
            return await self.ios_client.open_url(url)  # type: ignore[call-arg]
        except Exception as e:
            logger.error(f"Failed to open URL {url}: {e}")
            return False

    async def press_back(self) -> bool:
        """iOS doesn't have a back button - swipe from left edge."""
        try:
            # Simulate back gesture by swiping from left edge
            start = CoordinatesSelectorRequest(x=10, y=self.device_height // 4)
            end = CoordinatesSelectorRequest(x=300, y=self.device_height // 4)
            result = await self.swipe(start, end, duration=300)
            return result is None
        except Exception as e:
            logger.error(f"Failed to press back: {e}")
            return False

    async def press_home(self) -> bool:
        """Press the home button."""
        try:
            if self._is_idb:
                return await self.ios_client.button(button_type=HIDButtonType.HOME)  # type: ignore[call-arg, union-attr]
            else:
                return await self.ios_client.home()  # type: ignore[union-attr]
        except Exception as e:
            logger.error(f"Failed to press home: {e}")
            return False

    async def press_enter(self) -> bool:
        """Press the enter/return key."""
        try:
            await self.ios_client.key(40)  # type: ignore[call-arg]
            return True
        except Exception as e:
            logger.error(f"Failed to press enter: {e}")
            return False

    async def get_ui_hierarchy(self) -> list[dict]:
        """Get UI hierarchy using IDB accessibility info or WDA source."""
        try:
            accessibility_info = await asyncio.wait_for(
                self.ios_client.describe_all(), timeout=40.0
            )
            if accessibility_info is None:
                logger.warning("Accessibility info returned None")
                return []

            hierarchy = self._process_flat_ios_hierarchy(accessibility_info)
            return hierarchy
        except TimeoutError:
            logger.error("Timeout waiting for UI hierarchy (40s)")
            return []
        except Exception as e:
            logger.error(f"Failed to get UI hierarchy: {e}")
            return []

    def _process_flat_ios_hierarchy(self, accessibility_data: list[dict]) -> list[dict]:
        """
        Process iOS accessibility info into our standard format.

        Recursively flattens nested children into a flat list.
        """
        elements: list[dict] = []
        self._flatten_hierarchy(accessibility_data, elements)
        return elements

    def _flatten_hierarchy(self, nodes: list[dict], elements: list[dict]) -> None:
        """Recursively flatten the hierarchy tree into a flat list."""
        for node in nodes:
            if not isinstance(node, dict):
                continue

            # Extract element info
            element = {
                "type": node.get("type", ""),
                "value": node.get("AXValue") or node.get("value", ""),
                "label": node.get("AXLabel") or node.get("label", ""),
                "frame": node.get("frame", {}),
                "enabled": node.get("enabled", False),
                "visible": True,
            }

            # Add bounds if frame is available
            if "frame" in node and isinstance(node["frame"], dict):
                frame = node["frame"]
                if all(k in frame for k in ["x", "y", "width", "height"]):
                    element["bounds"] = (
                        f"[{int(frame['x'])},{int(frame['y'])}]"
                        f"[{int(frame['x'] + frame['width'])},{int(frame['y'] + frame['height'])}]"
                    )

            elements.append(element)

            # Recursively process children
            children = node.get("children", [])
            if children:
                self._flatten_hierarchy(children, elements)

    def find_element(
        self,
        ui_hierarchy: list[dict],
        resource_id: str | None = None,
        text: str | None = None,
        index: int = 0,
    ) -> tuple[dict | None, Bounds | None, str | None]:
        """Find a UI element in the iOS hierarchy."""
        if not resource_id and not text:
            return None, None, "No resource_id or text provided"

        matches = []
        for element in ui_hierarchy:
            # iOS doesn't have resource-id, so we match on type if provided as resource_id
            if resource_id and element.get("type") == resource_id:
                matches.append(element)
            # Match on value or label for text
            elif text and (element.get("value") == text or element.get("label") == text):
                matches.append(element)

        if not matches:
            criteria = f"type='{resource_id}'" if resource_id else f"text='{text}'"
            return None, None, f"No element found with {criteria}"

        if index >= len(matches):
            criteria = f"type='{resource_id}'" if resource_id else f"text='{text}'"
            return (
                None,
                None,
                f"Index {index} out of range for {criteria} (found {len(matches)} matches)",
            )

        element = matches[index]
        bounds = self._extract_bounds(element)

        return element, bounds, None

    def _extract_bounds(self, element: dict) -> Bounds | None:
        """Extract bounds from an iOS UI element."""
        bounds_str = element.get("bounds")
        if not bounds_str or not isinstance(bounds_str, str):
            return None

        try:
            # Parse bounds string like "[x1,y1][x2,y2]"
            match = re.match(r"\[(\d+),(\d+)\]\[(\d+),(\d+)\]", bounds_str)
            if match:
                return Bounds(
                    x1=int(match.group(1)),
                    y1=int(match.group(2)),
                    x2=int(match.group(3)),
                    y2=int(match.group(4)),
                )
        except (ValueError, IndexError):
            return None

        return None

    async def erase_text(self, nb_chars: int | None = None) -> bool:
        """Erase text by sending delete key presses."""
        try:
            if nb_chars is None:
                nb_chars = 50  # Default to erasing 50 characters
            # iOS delete key code is 42 (HID keyboard delete)
            for _ in range(nb_chars):
                await self.ios_client.key(42)  # type: ignore[call-arg]
            return True
        except Exception as e:
            logger.error(f"Failed to erase text: {e}")
            return False

    async def app_current(self) -> IOSAppInfo | None:
        """Get information about the currently active app."""
        try:
            return await self.ios_client.app_current()
        except Exception as e:
            logger.error(f"Failed to get current app: {e}")
            return None

    async def cleanup(self) -> None:
        """Cleanup iOS controller resources."""
        logger.debug("iOS controller cleanup")
        await self.ios_client.cleanup()

    def get_compressed_b64_screenshot(self, image_base64: str, quality: int = 50) -> str:
        if image_base64.startswith("data:image"):
            image_base64 = image_base64.split(",")[1]

        image_data = base64.b64decode(image_base64)
        image = Image.open(BytesIO(image_data))

        # Convert RGBA to RGB if image has alpha channel (PNG transparency)
        if image.mode in ("RGBA", "LA", "P"):
            rgb_image = Image.new("RGB", image.size, (255, 255, 255))
            rgb_image.paste(image, mask=image.split()[-1] if image.mode == "RGBA" else None)
            image = rgb_image

        compressed_io = BytesIO()
        image.save(compressed_io, format="JPEG", quality=quality, optimize=True)

        compressed_base64 = base64.b64encode(compressed_io.getvalue()).decode("utf-8")
        return compressed_base64

    async def start_video_recording(
        self,
        max_duration_seconds: int = DEFAULT_MAX_DURATION_SECONDS,
    ) -> VideoRecordingResult:
        """Start screen recording on iOS device/simulator using idb record-video."""
        device_id = self.device_id

        if has_active_session(device_id):
            return VideoRecordingResult(
                success=False,
                message=f"Recording already in progress for device {device_id}",
            )

        try:
            temp_dir = tempfile.mkdtemp(prefix="mobile_use_video_")
            video_path = Path(temp_dir) / "recording.mp4"

            cmd = ["idb", "record-video", "--udid", device_id, str(video_path)]

            process = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

            session = RecordingSession(
                device_id=device_id,
                start_time=time.time(),
                process=process,
                local_video_path=video_path,
            )
            set_active_session(device_id, session)

            logger.info(f"Started iOS screen recording on {device_id}")
            return VideoRecordingResult(
                success=True,
                message=f"Recording started (max {max_duration_seconds}s).",
            )

        except Exception as e:
            logger.error(f"Failed to start iOS recording: {e}")
            return VideoRecordingResult(
                success=False,
                message=f"Failed to start recording: {e}",
            )

    async def stop_video_recording(self) -> VideoRecordingResult:
        """Stop iOS recording and retrieve the video file."""
        device_id = self.device_id
        session = get_active_session(device_id)

        if not session:
            return VideoRecordingResult(
                success=False,
                message=f"No active recording for device {device_id}",
            )

        try:
            process = session.process
            if process is not None:
                try:
                    process.terminate()
                    await asyncio.wait_for(process.wait(), timeout=10.0)
                except TimeoutError:
                    process.kill()
                    await process.wait()

            await asyncio.sleep(VIDEO_READY_DELAY_SECONDS)

            local_path = session.local_video_path
            remove_active_session(device_id)

            duration = time.time() - session.start_time
            logger.info(f"Stopped iOS recording after {duration:.1f}s, saved to {local_path}")

            if local_path and local_path.exists():
                return VideoRecordingResult(
                    success=True,
                    message=f"Recording stopped after {duration:.1f}s",
                    video_path=local_path,
                )
            else:
                return VideoRecordingResult(
                    success=False,
                    message="Recording stopped but video file not found",
                )

        except Exception as e:
            logger.error(f"Failed to stop iOS recording: {e}")
            remove_active_session(device_id)
            return VideoRecordingResult(
                success=False,
                message=f"Failed to stop recording: {e}",
            )



import os
from pathlib import Path

from pydantic import ValidationError

from minitap.mobile_use.config import LLMConfig, deep_merge_llm_config, get_default_llm_config
from minitap.mobile_use.utils.file import load_jsonc
from minitap.mobile_use.utils.logger import get_logger

logger = get_logger(__name__)


def load_llm_config_override(path: Path) -> LLMConfig:
    default_config = get_default_llm_config()

    override_config_dict = {}
    if os.path.exists(path):
        logger.info(f"Loading custom LLM config from {path.resolve()}...")
        with open(path) as f:
            override_config_dict = load_jsonc(f)
    else:
        logger.warning("Custom LLM config not found - using the default config")

    try:
        return deep_merge_llm_config(default_config, override_config_dict)
    except ValidationError as e:
        logger.error(f"Invalid LLM config: {e}")
        logger.info("Falling back to default config")
        return default_config



from minitap.mobile_use.context import DevicePlatform, MobileUseContext
from minitap.mobile_use.controllers.android_controller import AndroidDeviceController
from minitap.mobile_use.controllers.device_controller import MobileDeviceController
from minitap.mobile_use.controllers.ios_controller import iOSDeviceController
from minitap.mobile_use.utils.logger import get_logger

logger = get_logger(__name__)


def create_device_controller(ctx: MobileUseContext) -> MobileDeviceController:
    platform = ctx.device.mobile_platform

    if platform == DevicePlatform.ANDROID:
        # Use Limrun controller if available (for cloud devices)
        if ctx.limrun_android_controller is not None:
            logger.info(f"Using LimrunAndroidController for device {ctx.device.device_id}")
            return ctx.limrun_android_controller

        if ctx.adb_client is None:
            raise ValueError("ADB client not initialized for Android device")

        if ctx.ui_adb_client is None:
            raise ValueError("UIAutomator client not initialized for Android device")

        logger.info(f"Creating Android controller for device {ctx.device.device_id}")
        return AndroidDeviceController(
            device_id=ctx.device.device_id,
            adb_client=ctx.adb_client,
            ui_adb_client=ctx.ui_adb_client,
            device_width=ctx.device.device_width,
            device_height=ctx.device.device_height,
        )

    elif platform == DevicePlatform.IOS:
        if ctx.ios_client is None:
            raise ValueError("iOS client not initialized for iOS device")

        logger.info(f"Creating iOS controller for device {ctx.device.device_id}")
        return iOSDeviceController(
            ios_client=ctx.ios_client,
            device_id=ctx.device.device_id,
            device_width=ctx.device.device_width,
            device_height=ctx.device.device_height,
        )

    else:
        raise ValueError(f"Unsupported platform: {platform}")


def get_controller(ctx: MobileUseContext) -> MobileDeviceController:
    return create_device_controller(ctx)



import asyncio
from functools import wraps
from typing import Any

from appium.options.common.base import AppiumOptions
from appium.webdriver.webdriver import WebDriver
from selenium.webdriver.common.actions import interaction
from selenium.webdriver.common.actions.action_builder import ActionBuilder
from selenium.webdriver.common.actions.pointer_input import PointerInput

from minitap.mobile_use.clients.idb_client import IOSAppInfo
from minitap.mobile_use.clients.ios_client_config import BrowserStackClientConfig
from minitap.mobile_use.utils.logger import get_logger

logger = get_logger(__name__)

BROWSERSTACK_HUB_URL = "https://hub-cloud.browserstack.com/wd/hub"


def with_browserstack_client(func):
    """Decorator to handle BrowserStack client error handling.

    Note: Function must have None or bool in return type for error fallback.
    """

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        method_name = func.__name__
        try:
            logger.debug(f"Executing BrowserStack operation: {method_name}...")
            result = await func(self, *args, **kwargs)
            logger.debug(f"{method_name} completed successfully")
            return result
        except Exception as e:
            logger.error(f"Failed to {method_name}: {e}")
            import traceback

            logger.debug(f"Traceback: {traceback.format_exc()}")

            return_type = func.__annotations__.get("return")
            if return_type is bool:
                return False
            return None

    return wrapper


class BrowserStackClientWrapper:
    """Wrapper around Appium WebDriver for BrowserStack iOS device automation.

    This wrapper provides an interface similar to IdbClientWrapper and WdaClientWrapper
    but uses BrowserStack's cloud infrastructure for physical iOS device automation.

    BrowserStack is used for:
    - Cloud-based physical iOS devices
    - CI/CD pipelines requiring real device testing
    - Cross-device testing without local hardware

    Prerequisites:
        1. BrowserStack account with App Automate access
        2. Valid username and access_key
        3. App uploaded to BrowserStack (app_url)

    Example:
        config = BrowserStackClientConfig(
            username="your_username",
            access_key="your_access_key",
            device_name="iPhone 14",
            platform_version="16",
            app_url="bs://your_app_hash",
        )
        wrapper = BrowserStackClientWrapper(config=config)
        await wrapper.init_client()
        await wrapper.tap(100, 200)
        await wrapper.cleanup()

        # Using context manager
        async with BrowserStackClientWrapper(config=config) as wrapper:
            await wrapper.tap(100, 200)
            screenshot = await wrapper.screenshot()
    """

    def __init__(self, config: BrowserStackClientConfig):
        """Initialize the BrowserStack client wrapper.

        Args:
            config: BrowserStack configuration with credentials and device settings
        """
        self.config = config
        self._driver: WebDriver | None = None

    async def init_client(self) -> bool:
        """Initialize the Appium WebDriver session on BrowserStack.

        Returns:
            True if session created successfully, False otherwise
        """
        try:
            logger.info(
                f"Creating BrowserStack session for {self.config.device_name} "
                f"(iOS {self.config.platform_version})"
            )

            options = AppiumOptions()

            options.set_capability("platformName", "iOS")
            options.set_capability("appium:deviceName", self.config.device_name)
            options.set_capability("appium:platformVersion", self.config.platform_version)
            options.set_capability("appium:automationName", "XCUITest")
            options.set_capability("appium:app", self.config.app_url)

            bstack_options: dict[str, Any] = {
                "userName": self.config.username,
                "accessKey": self.config.access_key.get_secret_value(),
                "buildName": self.config.build_name or "mobile-use-session",
                "sessionName": self.config.session_name or "BrowserStack Session",
                "debug": True,
            }

            if self.config.project_name:
                bstack_options["projectName"] = self.config.project_name

            options.set_capability("bstack:options", bstack_options)

            hub_url = self.config.hub_url or BROWSERSTACK_HUB_URL

            self._driver = await asyncio.to_thread(
                WebDriver,
                command_executor=hub_url,
                options=options,
            )

            if self._driver:
                session_id = self._driver.session_id
                logger.info(f"BrowserStack session created successfully. Session ID: {session_id}")
                logger.info(
                    f"View session: https://app-automate.browserstack.com/dashboard/v2/sessions/{session_id}"
                )

            return True

        except Exception as e:
            logger.error(f"Failed to create BrowserStack session: {e}")
            self._driver = None
            return False

    async def cleanup(self) -> None:
        """Clean up BrowserStack session and quit the driver."""
        if self._driver is not None:
            try:
                logger.info("Ending BrowserStack session")
                await asyncio.to_thread(self._driver.quit)
            except Exception as e:
                logger.debug(f"Error ending BrowserStack session: {e}")
            finally:
                self._driver = None

        logger.debug("BrowserStack client cleanup completed")

    async def __aenter__(self):
        """Async context manager entry."""
        if not await self.init_client():
            raise RuntimeError("Failed to create BrowserStack session")
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.cleanup()
        return False

    def _ensure_driver(self) -> WebDriver:
        """Ensure a valid WebDriver session exists.

        Returns:
            The WebDriver instance

        Raises:
            RuntimeError: If no driver is available
        """
        if self._driver is None:
            raise RuntimeError(
                "BrowserStack session not initialized. "
                "Call init_client() first or use as context manager."
            )
        return self._driver

    @with_browserstack_client
    async def tap(self, x: int, y: int, duration: float | None = None) -> bool:
        """Tap at the specified coordinates.

        Args:
            x: X coordinate
            y: Y coordinate
            duration: Optional tap duration in seconds (for long press)

        Returns:
            True if tap succeeded, False otherwise
        """
        driver = self._ensure_driver()

        def perform_tap():
            finger = PointerInput(interaction.POINTER_TOUCH, "finger")
            actions = ActionBuilder(driver, mouse=finger)
            actions.pointer_action.move_to_location(x, y)
            actions.pointer_action.pointer_down()
            if duration:
                actions.pointer_action.pause(duration)
            actions.pointer_action.pointer_up()
            actions.perform()

        await asyncio.to_thread(perform_tap)
        return True

    @with_browserstack_client
    async def swipe(
        self,
        x_start: int,
        y_start: int,
        x_end: int,
        y_end: int,
        duration: float | None = None,
    ) -> bool:
        """Swipe from start coordinates to end coordinates.

        Args:
            x_start: Starting X coordinate
            y_start: Starting Y coordinate
            x_end: Ending X coordinate
            y_end: Ending Y coordinate
            duration: Optional swipe duration in seconds

        Returns:
            True if swipe succeeded, False otherwise
        """
        driver = self._ensure_driver()

        swipe_duration = duration or 0.5

        def perform_swipe():
            finger = PointerInput(interaction.POINTER_TOUCH, "finger")
            actions = ActionBuilder(driver, mouse=finger)
            actions.pointer_action.move_to_location(x_start, y_start)
            actions.pointer_action.pointer_down()
            actions.pointer_action.pause(swipe_duration)
            actions.pointer_action.move_to_location(x_end, y_end)
            actions.pointer_action.pointer_up()
            actions.perform()

        await asyncio.to_thread(perform_swipe)
        return True

    @with_browserstack_client
    async def screenshot(self, output_path: str | None = None) -> bytes | None:
        """Take a screenshot and return raw image data.

        Args:
            output_path: Optional path to save the screenshot

        Returns:
            Raw image data (PNG bytes) or None on failure
        """
        driver = self._ensure_driver()

        screenshot_base64 = await asyncio.to_thread(driver.get_screenshot_as_base64)

        import base64

        screenshot_data = base64.b64decode(screenshot_base64)

        if output_path:
            with open(output_path, "wb") as f:
                f.write(screenshot_data)

        return screenshot_data

    @with_browserstack_client
    async def launch(
        self,
        bundle_id: str,
        args: list[str] | None = None,
        env: dict[str, str] | None = None,
    ) -> bool:
        """Launch an application by bundle ID.

        Args:
            bundle_id: The bundle identifier of the app to launch
            args: Optional list of arguments to pass to the app (not supported on BrowserStack)
            env: Optional environment variables for the app (not supported on BrowserStack)

        Returns:
            True if launch succeeded, False otherwise
        """
        driver = self._ensure_driver()

        if args or env:
            logger.warning(
                "BrowserStack does not support app launch arguments or environment variables"
            )

        script = "mobile: launchApp"
        params = {"bundleId": bundle_id}

        await asyncio.to_thread(driver.execute_script, script, params)
        return True

    @with_browserstack_client
    async def terminate(self, bundle_id: str) -> bool:
        """Terminate an application by bundle ID.

        Args:
            bundle_id: The bundle identifier of the app to terminate

        Returns:
            True if termination succeeded, False otherwise
        """
        driver = self._ensure_driver()

        script = "mobile: terminateApp"
        params = {"bundleId": bundle_id}

        await asyncio.to_thread(driver.execute_script, script, params)
        return True

    @with_browserstack_client
    async def text(self, text: str) -> bool:
        """Type text using the keyboard.

        Args:
            text: The text to type

        Returns:
            True if text input succeeded, False otherwise
        """
        driver = self._ensure_driver()

        active_element = await asyncio.to_thread(lambda: driver.switch_to.active_element)
        await asyncio.to_thread(active_element.send_keys, text)
        return True

    @with_browserstack_client
    async def open_url(self, url: str) -> bool:
        """Open a URL on the device.

        Args:
            url: The URL to open

        Returns:
            True if URL opened successfully, False otherwise
        """
        driver = self._ensure_driver()

        await asyncio.to_thread(driver.get, url)
        return True

    @with_browserstack_client
    async def key(self, key_code: int) -> bool:
        """Send a key press.

        Note: Limited key support on BrowserStack/Appium.
        For delete (key_code=42), we send a backspace.

        Args:
            key_code: HID key code (42 = delete/backspace)

        Returns:
            True if key press succeeded, False otherwise
        """
        driver = self._ensure_driver()

        if key_code == 42:  # Delete/backspace
            active_element = await asyncio.to_thread(lambda: driver.switch_to.active_element)
            current_text = await asyncio.to_thread(lambda: active_element.text)
            if current_text:
                await asyncio.to_thread(active_element.clear)
                await asyncio.to_thread(active_element.send_keys, current_text[:-1])
        return True

    @with_browserstack_client
    async def button(self, button_type: Any) -> bool:
        """Press a hardware button (compatible with IDB's HIDButtonType).

        Args:
            button_type: Button type (HIDButtonType.HOME, etc.)

        Returns:
            True if button press succeeded, False otherwise
        """
        driver = self._ensure_driver()

        button_name = getattr(button_type, "name", str(button_type)).lower()

        if button_name == "home":
            script = "mobile: pressButton"
            params = {"name": "home"}
            await asyncio.to_thread(driver.execute_script, script, params)
        elif button_name in ("volume_up", "volumeup"):
            script = "mobile: pressButton"
            params = {"name": "volumeUp"}
            await asyncio.to_thread(driver.execute_script, script, params)
        elif button_name in ("volume_down", "volumedown"):
            script = "mobile: pressButton"
            params = {"name": "volumeDown"}
            await asyncio.to_thread(driver.execute_script, script, params)

        return True

    async def describe_all(self) -> list[dict[str, Any]] | None:
        """Get UI hierarchy as a flat list (compatible with IDB's describe_all).

        Returns:
            List of UI elements or None on error
        """
        try:
            driver = self._ensure_driver()
            page_source = await asyncio.to_thread(lambda: driver.page_source)
            if page_source is None:
                return None
            return self._parse_xml_to_elements(page_source)
        except Exception as e:
            logger.error(f"Failed to describe_all: {e}")
            return None

    def _parse_xml_to_elements(self, xml_source: str) -> list[dict[str, Any]]:
        """Parse Appium XML source into flat element list matching IDB format."""
        import xml.etree.ElementTree as ET

        elements = []
        try:
            root = ET.fromstring(xml_source)
            for elem in root.iter():
                if elem.tag == "AppiumAUT":
                    continue
                frame = {
                    "x": float(elem.get("x", 0)),
                    "y": float(elem.get("y", 0)),
                    "width": float(elem.get("width", 0)),
                    "height": float(elem.get("height", 0)),
                }
                element = {
                    "type": elem.get("type", elem.tag),
                    "value": elem.get("value", ""),
                    "label": elem.get("label", elem.get("name", "")),
                    "frame": frame,
                    "enabled": elem.get("enabled", "false").lower() == "true",
                    "visible": elem.get("visible", "true").lower() == "true",
                }
                elements.append(element)
        except ET.ParseError as e:
            logger.error(f"Failed to parse XML: {e}")
        return elements

    async def app_current(self) -> IOSAppInfo | None:
        """Get information about the currently active app.

        Note: BrowserStack doesn't support activeAppInfo script directly.
        Returns None as this feature is not available on BrowserStack.

        Returns:
            None (not supported on BrowserStack)
        """
        logger.debug("app_current is not supported on BrowserStack")
        return None

    async def install(self, app_path: str) -> list[Any]:
        """Install an app (not supported on BrowserStack - apps must be pre-uploaded).

        Args:
            app_path: Path to the app (ignored)

        Returns:
            Empty list with warning
        """
        logger.warning(
            "App installation not supported on BrowserStack. "
            "Please upload your app to BrowserStack first and use the app_url in config."
        )
        return []
